import numpy as np
import os,json,sys
from typing import Dict, Sequence, Type, Callable, List, Optional
import torch
from torch import nn
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from modules.transformers import BertModel,BertConfig,BertTokenizer
from modules.MLP import MLP
from modules.Pooling import Pooling
from log import logging
logger=logging.getLogger(__name__)

'''
heads下的所有modelhead返回的都是logits，骨架中的module和具体的head model都用head进行load和save .head的输入是
就是input_ids这些特征，因此module和heads其实是绑在一起的

在语义相似度匹配任务中，骨架的设定是BiEncoders，heads的设定主要是不同的loss model，包括（softmaxloss，
ContrastiveLoss等）

每一个heads的forward返回的都是logits，比如在SoftmaxLossHead中的logits就是两个句子Pooling后的两个sentence_embeddings拼接后
传入到最终的classifier得到的logits，由于是两个句子，因此为了简便，forward的参数不是input_ids1和input_ids2这样区分
，而是[{'input_ids':tensor,'attention_mask':tensor,'token_type_ids':tensor},{'input_ids':tensor,'attention_mask':tensor,'token_type_ids':tensor}]
'''

class SoftmaxLossHead(nn.Module):
    """
    model应该根据对应的features给出句子的embedding，
    """
    def __init__(self,
                 model_path,
                 num_labels: int,
                 concatenation_sent_rep: bool = True,
                 concatenation_sent_difference: bool = True,
                 concatenation_sent_multiplication: bool = False,
                 pooling_mode_cls_token: bool = False,
                 pooling_mode_max_tokens: bool = False,
                 pooling_mode_mean_tokens: bool = True,
                 pooling_mode_mean_sqrt_len_tokens: bool = False,
                 pooling_mode_mean_last_2_tokens: bool = False,
                 pooling_mode_mean_first_last_tokens: bool = False,
                 is_finetune=False,
                 ):
        super(SoftmaxLossHead, self).__init__()
        self.num_labels = num_labels
        self.concatenation_sent_rep = concatenation_sent_rep
        self.concatenation_sent_difference = concatenation_sent_difference
        self.concatenation_sent_multiplication = concatenation_sent_multiplication
        self.pooling_mode_cls_token = pooling_mode_cls_token
        self.pooling_mode_max_tokens = pooling_mode_max_tokens
        self.pooling_mode_mean_tokens = pooling_mode_mean_tokens
        self.pooling_mode_mean_sqrt_len_tokens = pooling_mode_mean_sqrt_len_tokens
        self.pooling_mode_mean_last_2_tokens = pooling_mode_mean_last_2_tokens
        self.pooling_mode_mean_first_last_tokens = pooling_mode_mean_first_last_tokens

        self.num_vectors_concatenated = 0
        if concatenation_sent_rep:
            self.num_vectors_concatenated += 2
        if concatenation_sent_difference:
            self.num_vectors_concatenated += 1
        if concatenation_sent_multiplication:
            self.num_vectors_concatenated += 1

        self.config_keys = ['concatenation_sent_rep','concatenation_sent_difference','concatenation_sent_multiplication',
                            'sentence_embedding_dimension', 'pooling_mode_cls_token', 'pooling_mode_mean_tokens',
                            'pooling_mode_max_tokens', 'pooling_mode_mean_sqrt_len_tokens',
                            'pooling_mode_mean_last_2_tokens','pooling_mode_mean_first_last_tokens']
        self.pooling_config = {'pooling_mode_cls_token':self.pooling_mode_cls_token,
                                'pooling_mode_max_tokens':self.pooling_mode_max_tokens,
                                'pooling_mode_mean_tokens':self.pooling_mode_mean_tokens,
                                'pooling_mode_mean_sqrt_len_tokens':self.pooling_mode_mean_sqrt_len_tokens,
                                'pooling_mode_mean_last_2_tokens':self.pooling_mode_mean_last_2_tokens,
                                'pooling_mode_mean_first_last_tokens':self.pooling_mode_mean_first_last_tokens
                                }

        if is_finetune==False:
            logger.info("Loading model from {}, which is from huggingface model".format(model_path))
            self.load_huggingface_model(bert_model_path=model_path)
        else:
            self.load_finetuned_model(model_path=model_path)
            logger.info("Loading model from {}, which has been finetuned.".format(model_path))
        
        #logger.info("Pooling config : {}".format(self.pooling_config))
        logger.info("Softmax loss: #Vectors concatenated: {}".format(self.num_vectors_concatenated))
        logger.info("Pooling policy is ")
        logger.info("After pooling, each sentence embedding has dim: {}".format(self.pooling_layer.pooling_output_dimension))
    def load_huggingface_model(self,bert_model_path):
        self.config=BertConfig.from_pretrained(bert_model_path)
        self.bert=BertModel.from_pretrained(bert_model_path)
        self.tokenizer=BertTokenizer.from_pretrained(bert_model_path)
        self.pooling_config.update({'word_embedding_dimension':self.bert.config.hidden_size})
        self.pooling_layer=Pooling(**self.pooling_config)
        self.head_layer=MLP(in_features=self.num_vectors_concatenated*self.pooling_layer.pooling_output_dimension,out_features=self.num_labels)

    def load_finetuned_model(self,model_path):
        bert_save_path=os.path.join(model_path,"BERT")#save的时候将BERT保存在model_path下的BERT文件夹中
        self.config=BertConfig.from_pretrained(bert_save_path)
        self.bert=BertModel.from_pretrained(bert_save_path)
        self.tokenizer=BertTokenizer.from_pretrained(bert_save_path)

        pooling_save_path=os.path.join(model_path,'Pooling')
        self.pooling_layer.load(pooling_save_path)

        head_save_path=os.path.join(model_path,'MLP')
        self.head_layer=MLP.load(input_path=head_save_path)

    def save(self,output_path):
        bert_save_path=os.path.join(output_path,"BERT")
        self.bert.save_pretrained(bert_save_path,save_config=False)#下面已经save，不用save两次，虽然没什么影响
        self.config.save_pretrained(bert_save_path)
        self.tokenizer.save_pretrained(bert_save_path)

        pooling_save_path=os.path.join(output_path,'Pooling')
        self.pooling_layer.save(pooling_save_path)#主要是保存config.json，Pooling没有参数

        head_save_path=os.path.join(output_path,'MLP')
        self.head_layer.save(head_save_path)

    def forward(self,sentence_features_of_1,sentence_features_of_2=None,output_all_encoded_layers=False,encode_pattern=False):
        '''
        each_features like : {'input_ids':tensor,'attention_mask':tensor,'token_type_ids':tensor},
        input_ids.size()==attention_mask.size()==token_type_ids.size()==position_ids.size()==(batch_size,seq_length)
        label_ids.size()==(batch_size,)
        '''
        #只有在encode模式下的single_batch才是有意义的，不然如果不是encode模式，只传入一个句子，有没有标签，无法返回任何值
        single_batch=False
        if sentence_features_of_2 is None:
            single_batch=True
            try:
                assert encode_pattern
            except:
                raise Exception("只传入了一个batch的句子，然而又不是encode模式，函数无法执行")
            pair_sentence_features=[sentence_features_of_1]
        else:
            pair_sentence_features=[sentence_features_of_1,sentence_features_of_2]
        batch_size,seq_len_1=sentence_features_of_1['input_ids'].size()

        pair_sentence_embeddings=[]
        for sentence_features in pair_sentence_features:
            input_ids=sentence_features['input_ids']
            token_type_ids=sentence_features['token_type_ids']
            attention_mask=sentence_features['attention_mask']

            (sequence_outputs,pooler_output)=self.bert(input_ids=input_ids,
                                                    token_type_ids=token_type_ids,
                                                    attention_mask=attention_mask,
                                                    output_all_encoded_layers=output_all_encoded_layers)
            #要注意到sequence_output[0]与pooled_output的区别在于pooler_output是经过一层tanh的
            if output_all_encoded_layers:
                all_layer_embeddings=sequence_outputs
                token_embeddings=sequence_outputs[-1]
            else:
                all_layer_embeddings=None
                token_embeddings=sequence_outputs

            cls_token_embeddings=pooler_output
            sentence_embedding=self.pooling_layer(token_embeddings=token_embeddings,
                                                cls_token_embeddings=cls_token_embeddings,
                                                attention_mask=attention_mask,
                                                all_layer_embeddings=all_layer_embeddings)
            pair_sentence_embeddings.append(sentence_embedding)
            assert sentence_embedding.size()==(batch_size,self.pooling_layer.pooling_output_dimension)

        if single_batch:
            rep_a=pair_sentence_embeddings[0]
        else:
            rep_a,rep_b=pair_sentence_embeddings

        if encode_pattern == True:
            if single_batch:
                return rep_a
            else:
                return rep_a,rep_b
        try:
            assert sentence_features_of_2 is not None and len(pair_sentence_embeddings)==2
        except:
            raise Exception("encode pattern是False，那么第二个batch不能为空")

        assert rep_a.size()==(batch_size,self.pooling_layer.pooling_output_dimension)==rep_b.size()
        vectors_concat=[]
        if self.concatenation_sent_rep:
            vectors_concat.append(rep_a)
            vectors_concat.append(rep_b)
        if self.concatenation_sent_difference:
            vectors_concat.append(torch.abs(rep_a-rep_b))
        if self.concatenation_sent_multiplication:
            vectors_concat.append(rep_a*rep_b)
        
        concat_embeddings = torch.cat(vectors_concat, 1)
        assert concat_embeddings.size()==(batch_size,self.num_vectors_concatenated*self.pooling_layer.pooling_output_dimension)
        logits=self.head_layer(concat_embeddings)#(batch_size,num_labels)
        return logits

class STSHead(nn.Module):
    '''
    ClsHead不区分是单句子还是双句子，因为处理逻辑是一样的
    '''
    def __init__(self, model_path,
                num_labels,
                state_dict=None,
                is_finetune=False):
        super().__init__()
        self.num_labels=num_labels

        if is_finetune==False:
            logger.info("Loading model from {}, which is from huggingface model".format(model_path))
            self.load_huggingface_model(bert_model_path=model_path)
        else:
            self.load_finetuned_model(model_path=model_path)
            logger.info("Loading model from {}, which has been finetuned.".format(model_path))

        # bert_model_path=os.path.join(model_path,"BERT")#save的时候将BERT保存在model_path下的BERT文件夹中
        # self.config=BertConfig.from_pretrained(bert_model_path)
        # self.bert=BertModel.from_pretrained(bert_model_path)
        # self.tokenizer=BertTokenizer.from_pretrained(bert_model_path)

    def load_huggingface_model(self,bert_model_path):
        self.config=BertConfig.from_pretrained(bert_model_path)
        self.bert=BertModel.from_pretrained(bert_model_path)
        self.tokenizer=BertTokenizer.from_pretrained(bert_model_path)
        self.head_layer=MLP(in_features=self.config.hidden_size,out_features=self.num_labels)

    def load_finetuned_model(self,model_path):
        bert_save_path=os.path.join(model_path,"BERT")#save的时候将BERT保存在model_path下的BERT文件夹中
        self.config=BertConfig.from_pretrained(bert_save_path)
        self.bert=BertModel.from_pretrained(bert_save_path)
        self.tokenizer=BertTokenizer.from_pretrained(bert_save_path)

        head_save_path=os.path.join(model_path,'MLP')
        self.head_layer=MLP.load(input_path=head_save_path)

    def save(self,output_path):
        bert_save_path=os.path.join(output_path,"BERT")
        self.bert.save_pretrained(bert_save_path,save_config=False)#下面已经save，不用save两次，虽然没什么影响
        self.config.save_pretrained(bert_save_path)
        self.tokenizer.save_pretrained(bert_save_path)
        head_save_path=os.path.join(output_path,'MLP')
        self.head_layer.save(head_save_path)
    
    def forward(self,input_ids,attention_mask=None,token_type_ids=None,label_ids=None,output_all_encoded_layers=False):
        '''
        input_ids.size()==attention_mask.size()==token_type_ids.size()==position_ids.size()==(batch_size,seq_length)
        label_ids.size()==(batch_size,)
        '''
        (sequence_outputs,pooled_output)=self.bert(input_ids=input_ids,
                                                  token_type_ids=token_type_ids,
                                                   attention_mask=attention_mask,
                                                   output_all_encoded_layers=output_all_encoded_layers)
        #要注意到sequence_output[0]与pooled_output的区别在于pooled_output是经过一层tanh的
        assert len(pooled_output.size())==2 and pooled_output.size(1)==self.config.hidden_size
        logits=self.head_layer(pooled_output)#(batch_size,num_labels)
        return logits
        # predictions=torch.argmax(logits,dim=1)
        # if label_ids is not None:
        #     loss=nn.CrossEntropyLoss(reduction="mean")(input=logits,target=label_ids)
        #     accuracy=(predictions==label_ids).float().mean()
        #     return loss,accuracy
        # else:
        #     return logits,predictions


import numpy as np
import os,json,sys
from typing import Dict, Sequence, Type, Callable, List, Optional
import torch
from torch import nn
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from modules.transformers import BertModel,BertConfig,BertTokenizer
from modules.MLP import MLP
from modules.LSTM import LSTMLayer
from log import logging
logger=logging.getLogger(__name__)

'''
BertModel inherite BertForPretrainedModel inherite PreTrainedModel

In PreTrainedModel
@classmethod
def from_pretrained(cls,pretrained_model_name_or_path,*model_args,**kwargs):
    config=kwargs.pop('config',None)
    state_dict=kwargs.pop('state_dict',None)

    #The config will be loaded from pretrained_model_name_or_path/config.json if config not provided
    #instantiate model
    model=cls(config)
    if state_dict is None:
        state_dict=torch.load(pretrained_model_name_or_path/pytorch_model.bin,map_location='cpu')#也就是说模型最开始是加载到cpu上的
    missing_keys,unexpected_keys,error_msgs=[],[],[]

    def load(module: nn.Module, prefix=''):
        module._load_from_state_dict(state_dict,prefix,missing_keys,unexpected_keys,error_msgs)
        for name,child in module._modules.items():
            if child:
                load(child,prefix+name+'.')
    
    load(model,prefix='' if hasattr(model,'bert') else 'bert.')
    #prefix用来解决加载不同模型保存的pytorch_model.bin出现的keys不匹配问题
    if len(missing_keys) > 0:
        logger.info("Weights of {} not initialized from pretrained model: {}".format(
            model.__class__.__name__, missing_keys))
    if len(unexpected_keys) > 0:
        logger.info("Weights from pretrained model not used in {}: {}".format(
            model.__class__.__name__, unexpected_keys))
    return model
'''

class CRF(nn.Module):
    def __init__(self, num_tags, batch_first=True) -> None:
        super().__init__()
        self.num_tags = num_tags
        self.batch_first = batch_first
        #一定要有nn.Parameter，用Parameter修饰的tensor才可以被认为是参数,requires_grad=True
        self.start_transitions = nn.Parameter(torch.empty(num_tags))
        self.end_transitions = nn.Parameter(torch.empty(num_tags))
        self.transitions = nn.Parameter(torch.empty(num_tags,num_tags))
        self.init_matrix()

    def init_matrix(self):
        nn.init.uniform_(self.start_transitions, -0.1, 0.1)
        nn.init.uniform_(self.end_transitions, -0.1, 0.1)
        nn.init.uniform_(self.transitions, -0.1, 0.1)

    def get_config_dict(self):
        '''
        一定要有dict，这样才能初始化Model
        '''
        return {'num_tags': self.num_tags, 'batch_first':self.batch_first}

    def save(self,output_path):
        '''
        同时保存dict
        '''
        os.makedirs(output_path,exist_ok=True)
        with open(os.path.join(output_path, 'config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut)

        torch.save(self.state_dict(),os.path.join(output_path, 'pytorch_model.bin'))
    
    @staticmethod
    def load(input_path):
        with open(os.path.join(input_path, 'config.json')) as fIn:
            config = json.load(fIn)

        model = CRF(**config)
        model.load_state_dict(torch.load(os.path.join(input_path, 'pytorch_model.bin'), map_location=torch.device('cpu')))
        return model

    def compute_numerator(self,emissions, tags, mask):
        '''
        emissions.size()=(max_seq_len.batch_size,num_tags)
        tags.size()=(max_seq_len,batch_size)
        mask.size()=(max_seq_len,batch_size)

        when max_seq_len=4,batch_size=3,num_tags=6
        examples: emissions=tensor([[[ 0.5219,  0.8717,  0.5480, -0.8730, -0.1198, -1.0771],
         [ 0.0299, -1.0110, -1.5722, -0.0844,  3.0301, -0.7060],
         [-0.8239, -1.3775, -0.9329, -1.6733,  0.2417, -0.2807]],

        [[ 0.6835, -0.3147, -0.2730,  0.8898, -0.4520, -2.3399],
         [-0.6476,  0.6417, -0.5508,  0.0535, -0.9227,  1.1172],
         [-1.7558,  0.7489, -0.9195,  1.3225, -0.7673, -0.9033]],

        [[-0.2535, -1.0029,  0.4376, -1.4911,  0.0859, -0.4517],
         [ 0.4292, -1.2318,  2.2034, -1.6605, -0.8760, -0.1399],
         [-1.0145,  0.1310, -1.1153, -1.5226,  0.4480, -0.8239]],

        [[-0.6559, -0.6905, -0.2348, -0.5770, -1.1463, -0.3917],
         [-0.0180,  1.0983,  1.4752,  0.7994,  1.4708, -0.6190],
         [ 0.5745,  1.1589, -0.0876, -0.7465, -1.4070,  1.3370]]])

         tags=tensor([[2, 4, 5],
        [0, 2, 4],
        [4, 3, 1],
        [2, 4, 1]])

        mask=tensor([[1, 1, 1],
        [1, 1, 1],
        [1, 1, 0],
        [1, 1, 0]])
        也就是说第三个句子的实际长度是2，pad位置对应的tag是1
        '''
        assert emissions.dim() == 3 and tags.dim() == 2
        assert emissions.shape[:2] == tags.shape
        assert emissions.size(2) == self.num_tags
        assert mask.shape == tags.shape
        assert mask[0].all() # tests if all elements in this tensor is True，验证所有句子的第一个时间步必须是True

        seq_length, batch_size = tags.shape
        mask = mask.float()
        #tags[0].size()==(batch_size,)，代表的是batch内的所有句子的第一个单词对应的标注，也就是说三个句子的第一个单词对应的标注分别是2,4,5
        score=self.start_transitions[tags[0]]#也就是初始概率矩阵，即初始时刻观测到标注2或4或5的概率分别是多少
        score+=emissions[0,torch.arange(batch_size),tags[0]]#也就是第一个时间步，这三个句子分别观测到2,4,5的观测概率，分别是tensor([ 0.5480,  3.0301, -0.2807])

        #真实路径的分数等于这条路径上的发射分数+转移分数
        for i in range(1,seq_length):
            score+=self.transitions[tags[i-1],tags[i]]*mask[i]#这三个句子从第一个单词的标注转移到第二个单词的标注的转移概率
            score+=emissions[i,torch.arange(batch_size),tags[i]]*mask[i]
            #这三个句子在第i个时间步计算得到的前向概率

        seq_ends=mask.long().sum(0)-1#三个句子的实际长度 [4,4,2]
        last_time_step_tags=tags[seq_ends,torch.arange(batch_size)]#三个句子在最后一个单词的实际标注 [2,4,4]

        score+=self.end_transitions[last_time_step_tags]

        return score

    def compute_denominator(self,emissions,mask):
        '''
        emissions.size()=(max_seq_len.batch_size,num_tags)
        mask.size()=(max_seq_len,batch_size)
        '''
        assert emissions.dim() == 3 and mask.dim() == 2
        assert emissions.shape[:2] == mask.shape
        assert emissions.size(2) == self.num_tags
        mask=mask.bool()
        assert mask[0].all()

        seq_length = emissions.size(0)

        score=self.start_transitions+emissions[0]#所有句子第一个时间步的前向概率,shape==(batch_size,num_tags)

        for i in range(1, seq_length):
            broadcast_score = score.unsqueeze(2)#(batch_size,num_tags,1)
            broadcast_emissions = emissions[i].unsqueeze(1)#(batch_size,1,num_tags)

            next_score = broadcast_score + self.transitions + broadcast_emissions#(batch_size,num_tags,num_tags)

            next_score = torch.logsumexp(next_score, dim=1)#(batch_size,num_tags)

            score = torch.where(mask[i].unsqueeze(1), next_score, score)

        score += self.end_transitions#(batch_size,num_tags)

        return torch.logsumexp(score,dim=1)#(batch_size,)

    def forward(
            self,
            emissions: torch.Tensor,
            tags: torch.LongTensor,
            mask: Optional[torch.ByteTensor] = None,
            reduction: str = 'token_mean',
            neg_loglikelihood=True):

        if mask is None:
            mask = torch.ones_like(tags, dtype=torch.uint8)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            tags = tags.transpose(0, 1)
            mask = mask.transpose(0, 1)
        #print(emissions.size(),tags.size(),mask.size())
        numerator=self.compute_numerator(emissions,tags=tags,mask=mask)
        denominator=self.compute_denominator(emissions=emissions,mask=mask)

        log_likelihood=numerator-denominator
        if reduction=='sum':
            log_likelihood=log_likelihood.sum()
        elif reduction=='mean':
            log_likelihood=log_likelihood.mean()
        elif reduction=='token_mean':
            log_likelihood=log_likelihood.sum()/mask.float().sum()
        else:
            raise Exception("reduction not recognized!")

        if neg_loglikelihood:
            return -log_likelihood
        else:
            return log_likelihood

    def viterbi_decode(self,emissions,mask):
        if mask is None:
            mask = emissions.new_ones(emissions.shape[:2], dtype=torch.uint8)

        if self.batch_first:
            emissions = emissions.transpose(0, 1)
            mask = mask.transpose(0, 1)
        #emissions.size()==(max_seq_len,batch_size,num_tags)
        assert emissions.dim() == 3 and mask.dim() == 2
        assert emissions.shape[:2] == mask.shape
        assert emissions.size(2) == self.num_tags
        mask=mask.bool()
        assert mask[0].all()

        seq_length, batch_size = mask.shape

        score = self.start_transitions + emissions[0]#三个句子在第一个时间步的前向概率
        history = []
        
        for i in range(1, seq_length):
            broadcast_score = score.unsqueeze(2)#(batch_size,num_tags,1)
            broadcast_emissions = emissions[i].unsqueeze(1)#(batch_size,1,num_tags)

            next_score = broadcast_score + self.transitions + broadcast_emissions#(batch_size,num_tags,num_tags)

            next_score, indices = next_score.max(dim=1)
            score = torch.where(mask[i].unsqueeze(1), next_score, score)
            history.append(indices)

        score += self.end_transitions
        seq_ends = mask.long().sum(dim=0) - 1
        best_tags_list = []

        for idx in range(batch_size):
            _, best_last_tag = score[idx].max(dim=0)
            best_tags = [best_last_tag.item()]

            for hist in reversed(history[:seq_ends[idx]]):
                best_last_tag = hist[idx][best_tags[-1]]
                best_tags.append(best_last_tag.item())

            best_tags.reverse()
            best_tags_list.append(best_tags)
        #返回的batch_tags_list是长度是3的列表，第一个元素是一个list，长度和第一个句子的长度一致，每一个值是
        #label2id中的id，代表当前时间步的预测
        return best_tags_list

class NerHead(nn.Module):
    def __init__(self, model_path,
                num_labels,
                state_dict=None,
                is_finetune=False,
                use_bilstm=False,
                use_crf=False):
        super().__init__()
        self.num_labels=num_labels
        self.use_bilstm=use_bilstm
        self.use_crf=use_crf

        if is_finetune==False:
            logger.info("Loading model from {}, which is from huggingface model".format(model_path))
            self.load_huggingface_model(bert_model_path=model_path)
        else:
            self.load_finetuned_model(model_path=model_path)
            logger.info("Loading model from {}, which has been finetuned.".format(model_path))

        # bert_model_path=os.path.join(model_path,"BERT")#save的时候将BERT保存在model_path下的BERT文件夹中
        # self.config=BertConfig.from_pretrained(bert_model_path)
        # self.bert=BertModel.from_pretrained(bert_model_path)
        # self.tokenizer=BertTokenizer.from_pretrained(bert_model_path)

    def load_huggingface_model(self,bert_model_path):
        self.config=BertConfig.from_pretrained(bert_model_path)
        self.bert=BertModel.from_pretrained(bert_model_path)
        self.tokenizer=BertTokenizer.from_pretrained(bert_model_path)
        if self.use_bilstm:
            self.lstmLayer=LSTMLayer(input_size=self.config.hidden_size,hidden_size=self.config.hidden_size,num_layers=1,batch_first=True,bidirectional=True)
            self.outputLayer=MLP(in_features=self.config.hidden_size*2,out_features=self.num_labels,bias=True)
        else:
            self.outputLayer=MLP(in_features=self.config.hidden_size,out_features=self.num_labels,bias=True)
        if self.use_crf:
            self.crfLayer=CRF(num_tags=self.num_labels,batch_first=True)

    def load_finetuned_model(self,model_path):
        bert_save_path=os.path.join(model_path,"BERT")#save的时候将BERT保存在model_path下的BERT文件夹中
        self.config=BertConfig.from_pretrained(bert_save_path)
        self.bert=BertModel.from_pretrained(bert_save_path)
        self.tokenizer=BertTokenizer.from_pretrained(bert_save_path)

        if self.use_bilstm:
            bistm_save_path=os.path.join(model_path,'BiLSTM')
            self.lstmLayer=LSTMLayer.load(bistm_save_path)
            self.outputLayer=MLP.load(os.path.join(model_path,'BiLSTM','MLP'))
        else:
            head_save_path=os.path.join(model_path,'MLP')
            self.outputLayer=MLP.load(head_save_path)

        if self.use_crf:
            crf_save_path=os.path.join(model_path,'CRF')
            self.crfLayer=CRF.load(crf_save_path)

    def save(self,output_path):
        bert_save_path=os.path.join(output_path,"BERT")
        self.bert.save_pretrained(bert_save_path,save_config=False)#下面已经save，不用save两次，虽然没什么影响
        self.config.save_pretrained(bert_save_path)
        self.tokenizer.save_pretrained(bert_save_path)
        if self.use_bilstm:
            bilstm_save_path=os.path.join(output_path,'BiLSTM')
            self.lstmLayer.save(bilstm_save_path)
            self.outputLayer.save(os.path.join(output_path,'BiLSTM','MLP'))
        else:
            head_save_path=os.path.join(output_path,'MLP')
            self.outputLayer.save(head_save_path)
        if self.use_crf:
            crf_save_path=os.path.join(output_path,'CRF')
            self.crfLayer.save(crf_save_path)
    
    def forward(self,input_ids,attention_mask=None,token_type_ids=None,label_ids=None,output_all_encoded_layers=False):
        '''
        input_ids.size()==attention_mask.size()==token_type_ids.size()==position_ids.size()==(batch_size,seq_length)
        label_ids.size()==(batch_size,)
        '''
        batch_size,max_seq_len=input_ids.size()
        (sequence_outputs,pooled_output)=self.bert(input_ids=input_ids,
                                                  token_type_ids=token_type_ids,
                                                   attention_mask=attention_mask,
                                                   output_all_encoded_layers=output_all_encoded_layers)
        #要注意到sequence_output[0]与pooled_output的区别在于pooled_output是经过一层tanh的
        #assert len(pooled_output.size())==2 and pooled_output.size(1)==self.config.hidden_size
        if self.use_bilstm:
            sequence_outputs=self.lstmLayer(sequence_outputs,attention_mask=attention_mask)
            logits=self.outputLayer(sequence_outputs)
        else:
            logits=self.outputLayer(sequence_outputs)
            
        assert logits.size()==(batch_size,max_seq_len,self.num_labels)
        return logits
        
        # if self.use_crf:
        #     loss=self.crfLayer(emissions=sequence_outputs,tags=label_ids,mask=attention_mask)
        #     return loss
        # else:
        #     return logits



# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""PyTorch optimization for BERT model."""

import math,sys,os
from typing import Callable, Iterable, Tuple

import torch
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR
#sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from .log import logging
logger = logging.getLogger(__name__)


def get_constant_schedule(optimizer: Optimizer, last_epoch: int = -1):
    """
    Create a schedule with a constant learning rate, using the learning rate set in optimizer.

    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """
    return LambdaLR(optimizer, lambda _: 1, last_epoch=last_epoch)


def get_constant_schedule_with_warmup(optimizer: Optimizer, num_warmup_steps: int, last_epoch: int = -1):
    """
    Create a schedule with a constant learning rate preceded by a warmup period during which the learning rate
    increases linearly between 0 and the initial lr set in the optimizer.

    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1.0, num_warmup_steps))
        return 1.0

    return LambdaLR(optimizer, lr_lambda, last_epoch=last_epoch)


def get_linear_schedule_with_warmup(optimizer, num_warmup_steps, num_training_steps, last_epoch=-1):
    """
    Create a schedule with a learning rate that decreases linearly from the initial lr set in the optimizer to 0,
    after a warmup period during which it increases linearly from 0 to the initial lr set in the optimizer.

    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        return max(
            0.0, float(num_training_steps - current_step) / float(max(1, num_training_steps - num_warmup_steps))
        )

    return LambdaLR(optimizer, lr_lambda, last_epoch)


def get_cosine_schedule_with_warmup(
    optimizer: Optimizer, num_warmup_steps: int, num_training_steps: int, num_cycles: float = 0.5, last_epoch: int = -1
):
    """
    Create a schedule with a learning rate that decreases following the values of the cosine function between the
    initial lr set in the optimizer to 0, after a warmup period during which it increases linearly between 0 and the
    initial lr set in the optimizer.

    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        num_cycles (:obj:`float`, `optional`, defaults to 0.5):
            The number of waves in the cosine schedule (the defaults is to just decrease from the max value to 0
            following a half-cosine).
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * float(num_cycles) * 2.0 * progress)))

    return LambdaLR(optimizer, lr_lambda, last_epoch)


def get_cosine_with_hard_restarts_schedule_with_warmup(
    optimizer: Optimizer, num_warmup_steps: int, num_training_steps: int, num_cycles: int = 1, last_epoch: int = -1
):
    """
    Create a schedule with a learning rate that decreases following the values of the cosine function between the
    initial lr set in the optimizer to 0, with several hard restarts, after a warmup period during which it increases
    linearly between 0 and the initial lr set in the optimizer.

    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        num_cycles (:obj:`int`, `optional`, defaults to 1):
            The number of hard restarts to use.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.

    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.
    """

    def lr_lambda(current_step):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        progress = float(current_step - num_warmup_steps) / float(max(1, num_training_steps - num_warmup_steps))
        if progress >= 1.0:
            return 0.0
        return max(0.0, 0.5 * (1.0 + math.cos(math.pi * ((float(num_cycles) * progress) % 1.0))))

    return LambdaLR(optimizer, lr_lambda, last_epoch)


def get_polynomial_decay_schedule_with_warmup(
    optimizer, num_warmup_steps, num_training_steps, lr_end=1e-7, power=1.0, last_epoch=-1
):
    """
    Create a schedule with a learning rate that decreases as a polynomial decay
    from the initial lr set in the optimizer to end lr defined by `lr_end`,
    after a warmup period during which it increases linearly from 0 to the
    initial lr set in the optimizer.

    Args:
        optimizer (:class:`~torch.optim.Optimizer`):
            The optimizer for which to schedule the learning rate.
        num_warmup_steps (:obj:`int`):
            The number of steps for the warmup phase.
        num_training_steps (:obj:`int`):
            The total number of training steps.
        lr_end (:obj:`float`, `optional`, defaults to 1e-7):
            The end LR.
        power (:obj:`float`, `optional`, defaults to 1.0):
            Power factor.
        last_epoch (:obj:`int`, `optional`, defaults to -1):
            The index of the last epoch when resuming training.

    Note: `power` defaults to 1.0 as in the fairseq implementation, which in turn is
    based on the original BERT implementation at
    https://github.com/google-research/bert/blob/f39e881b169b9d53bea03d2d341b31707a6c052b/optimization.py#L37

    Return:
        :obj:`torch.optim.lr_scheduler.LambdaLR` with the appropriate schedule.

    """

    lr_init = optimizer.defaults["lr"]
    assert lr_init > lr_end, f"lr_end ({lr_end}) must be be smaller than initial lr ({lr_init})"

    def lr_lambda(current_step: int):
        if current_step < num_warmup_steps:
            return float(current_step) / float(max(1, num_warmup_steps))
        elif current_step > num_training_steps:
            return lr_end / lr_init  # as LambdaLR multiplies by lr_init
        else:
            lr_range = lr_init - lr_end
            decay_steps = num_training_steps - num_warmup_steps
            pct_remaining = 1 - (current_step - num_warmup_steps) / decay_steps
            decay = lr_range * pct_remaining ** power + lr_end
            return decay / lr_init  # as LambdaLR multiplies by lr_init

    return LambdaLR(optimizer, lr_lambda, last_epoch)


class AdamW(Optimizer):
    """
    Implements Adam algorithm with weight decay fix as introduced in
    `Decoupled Weight Decay Regularization <https://arxiv.org/abs/1711.05101>`__.

    Parameters:
        params (:obj:`Iterable[torch.nn.parameter.Parameter]`):
            Iterable of parameters to optimize or dictionaries defining parameter groups.
        lr (:obj:`float`, `optional`, defaults to 1e-3):
            The learning rate to use.
        betas (:obj:`Tuple[float,float]`, `optional`, defaults to (0.9, 0.999)):
            Adam's betas parameters (b1, b2).
        eps (:obj:`float`, `optional`, defaults to 1e-6):
            Adam's epsilon for numerical stability.
        weight_decay (:obj:`float`, `optional`, defaults to 0):
            Decoupled weight decay to apply.
        correct_bias (:obj:`bool`, `optional`, defaults to `True`):
            Whether ot not to correct bias in Adam (for instance, in Bert TF repository they use :obj:`False`).
    """

    def __init__(
        self,
        params: Iterable[torch.nn.parameter.Parameter],
        lr: float = 1e-3,
        betas: Tuple[float, float] = (0.9, 0.999),
        eps: float = 1e-6,
        weight_decay: float = 0.0,
        correct_bias: bool = True,
    ):
        if lr < 0.0:
            raise ValueError("Invalid learning rate: {} - should be >= 0.0".format(lr))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter: {} - should be in [0.0, 1.0[".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter: {} - should be in [0.0, 1.0[".format(betas[1]))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {} - should be >= 0.0".format(eps))
        defaults = dict(lr=lr, betas=betas, eps=eps, weight_decay=weight_decay, correct_bias=correct_bias)
        super().__init__(params, defaults)

    def step(self, closure: Callable = None):
        """
        Performs a single optimization step.

        Arguments:
            closure (:obj:`Callable`, `optional`): A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.is_sparse:
                    raise RuntimeError("Adam does not support sparse gradients, please consider SparseAdam instead")

                state = self.state[p]

                # State initialization
                if len(state) == 0:
                    state["step"] = 0
                    # Exponential moving average of gradient values
                    state["exp_avg"] = torch.zeros_like(p.data)
                    # Exponential moving average of squared gradient values
                    state["exp_avg_sq"] = torch.zeros_like(p.data)

                exp_avg, exp_avg_sq = state["exp_avg"], state["exp_avg_sq"]
                beta1, beta2 = group["betas"]

                state["step"] += 1

                # Decay the first and second moment running average coefficient
                # In-place operations to update the averages at the same time
                exp_avg.mul_(beta1).add_(grad, alpha=1.0 - beta1)
                exp_avg_sq.mul_(beta2).addcmul_(grad, grad, value=1.0 - beta2)
                denom = exp_avg_sq.sqrt().add_(group["eps"])

                step_size = group["lr"]
                if group["correct_bias"]:  # No bias correction for Bert
                    bias_correction1 = 1.0 - beta1 ** state["step"]
                    bias_correction2 = 1.0 - beta2 ** state["step"]
                    step_size = step_size * math.sqrt(bias_correction2) / bias_correction1

                p.data.addcdiv_(exp_avg, denom, value=-step_size)

                # Just adding the square of the weights to the loss function is *not*
                # the correct way of using L2 regularization/weight decay with Adam,
                # since that will interact with the m and v parameters in strange ways.
                #
                # Instead we want to decay the weights in a manner that doesn't interact
                # with the m/v parameters. This is equivalent to adding the square
                # of the weights to the loss with plain (non-momentum) SGD.
                # Add weight decay at the end (fixed version)
                if group["weight_decay"] > 0.0:
                    p.data.add_(p.data, alpha=-group["lr"] * group["weight_decay"])

        return loss


class Adafactor(Optimizer):
    """
    AdaFactor pytorch implementation can be used as a drop in replacement for Adam
    original fairseq code: https://github.com/pytorch/fairseq/blob/master/fairseq/optim/adafactor.py

    Paper: `Adafactor: Adaptive Learning Rates with Sublinear Memory Cost` https://arxiv.org/abs/1804.04235
    Note that this optimizer internally adjusts the learning rate depending on the *scale_parameter*, *relative_step* and
    *warmup_init* options. To use a manual (external) learning rate schedule you should set `scale_parameter=False` and `relative_step=False`.

    Arguments:
        params (:obj:`Iterable[torch.nn.parameter.Parameter]`):
            Iterable of parameters to optimize or dictionaries defining parameter groups.
        lr (:obj:`float`, `optional`):
            The external learning rate.
        eps (:obj:`Tuple[float, float]`, `optional`, defaults to (1e-30, 1e-3)):
            Regularization constants for square gradient and parameter scale respectively
        clip_threshold (:obj:`float`, `optional`, defaults 1.0):
            Threshold of root mean square of final gradient update
        decay_rate (:obj:`float`, `optional`, defaults to -0.8):
            Coefficient used to compute running averages of square
        beta1 (:obj:`float`, `optional`):
            Coefficient used for computing running averages of gradient
        weight_decay (:obj:`float`, `optional`, defaults to 0):
            Weight decay (L2 penalty)
        scale_parameter (:obj:`bool`, `optional`, defaults to :obj:`True`):
            If True, learning rate is scaled by root mean square
        relative_step (:obj:`bool`, `optional`, defaults to :obj:`True`):
            If True, time-dependent learning rate is computed instead of external learning rate
        warmup_init (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Time-dependent learning rate computation depends on whether warm-up initialization is being used

    This implementation handles low-precision (FP16, bfloat) values, but we have not thoroughly tested.

    Recommended T5 finetuning settings:
        - Scheduled LR warm-up to fixed LR
        - disable relative updates
        - use clip threshold: https://arxiv.org/abs/2004.14546

        Example::

            Adafactor(model.parameters(), lr=1e-3, relative_step=False, warmup_init=True)

        - Alternatively, relative_step with warmup_init can be used.
        - Training without LR warmup or clip threshold is not recommended. Additional optimizer operations like
          gradient clipping should not be used alongside Adafactor.

    Usage::

        # replace AdamW with Adafactor
        optimizer = Adafactor(
            model.parameters(),
            lr=1e-3,
            eps=(1e-30, 1e-3),
            clip_threshold=1.0,
            decay_rate=-0.8,
            beta1=None,
            weight_decay=0.0,
            relative_step=False,
            scale_parameter=False,
            warmup_init=False
        )
    """

    def __init__(
        self,
        params,
        lr=None,
        eps=(1e-30, 1e-3),
        clip_threshold=1.0,
        decay_rate=-0.8,
        beta1=None,
        weight_decay=0.0,
        scale_parameter=True,
        relative_step=True,
        warmup_init=False,
    ):
        if lr is not None and relative_step:
            raise ValueError("Cannot combine manual lr and relative_step options")
        if warmup_init and not relative_step:
            raise ValueError("warmup_init requires relative_step=True")

        defaults = dict(
            lr=lr,
            eps=eps,
            clip_threshold=clip_threshold,
            decay_rate=decay_rate,
            beta1=beta1,
            weight_decay=weight_decay,
            scale_parameter=scale_parameter,
            relative_step=relative_step,
            warmup_init=warmup_init,
        )
        super().__init__(params, defaults)

    @staticmethod
    def _get_lr(param_group, param_state):
        rel_step_sz = param_group["lr"]
        if param_group["relative_step"]:
            min_step = 1e-6 * param_state["step"] if param_group["warmup_init"] else 1e-2
            rel_step_sz = min(min_step, 1.0 / math.sqrt(param_state["step"]))
        param_scale = 1.0
        if param_group["scale_parameter"]:
            param_scale = max(param_group["eps"][1], param_state["RMS"])
        return param_scale * rel_step_sz

    @staticmethod
    def _get_options(param_group, param_shape):
        factored = len(param_shape) >= 2
        use_first_moment = param_group["beta1"] is not None
        return factored, use_first_moment

    @staticmethod
    def _rms(tensor):
        return tensor.norm(2) / (tensor.numel() ** 0.5)

    @staticmethod
    def _approx_sq_grad(exp_avg_sq_row, exp_avg_sq_col):
        r_factor = (exp_avg_sq_row / exp_avg_sq_row.mean(dim=-1, keepdim=True)).rsqrt_()
        c_factor = exp_avg_sq_col.rsqrt()
        return torch.mm(r_factor.unsqueeze(-1), c_factor.unsqueeze(0))

    def step(self, closure=None):
        """Performs a single optimization step.
        Arguments:
            closure (callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                grad = p.grad.data
                if grad.dtype in {torch.float16, torch.bfloat16}:
                    grad = grad.float()
                if grad.is_sparse:
                    raise RuntimeError("Adafactor does not support sparse gradients.")

                state = self.state[p]
                grad_shape = grad.shape

                factored, use_first_moment = self._get_options(group, grad_shape)
                # State Initialization
                if len(state) == 0:
                    state["step"] = 0

                    if use_first_moment:
                        # Exponential moving average of gradient values
                        state["exp_avg"] = torch.zeros_like(grad)
                    if factored:
                        state["exp_avg_sq_row"] = torch.zeros(grad_shape[:-1]).to(grad)
                        state["exp_avg_sq_col"] = torch.zeros(grad_shape[:-2] + grad_shape[-1:]).to(grad)
                    else:
                        state["exp_avg_sq"] = torch.zeros_like(grad)

                    state["RMS"] = 0
                else:
                    if use_first_moment:
                        state["exp_avg"] = state["exp_avg"].to(grad)
                    if factored:
                        state["exp_avg_sq_row"] = state["exp_avg_sq_row"].to(grad)
                        state["exp_avg_sq_col"] = state["exp_avg_sq_col"].to(grad)
                    else:
                        state["exp_avg_sq"] = state["exp_avg_sq"].to(grad)

                p_data_fp32 = p.data
                if p.data.dtype in {torch.float16, torch.bfloat16}:
                    p_data_fp32 = p_data_fp32.float()

                state["step"] += 1
                state["RMS"] = self._rms(p_data_fp32)
                group["lr"] = self._get_lr(group, state)

                beta2t = 1.0 - math.pow(state["step"], group["decay_rate"])
                update = (grad ** 2) + group["eps"][0]
                if factored:
                    exp_avg_sq_row = state["exp_avg_sq_row"]
                    exp_avg_sq_col = state["exp_avg_sq_col"]

                    exp_avg_sq_row.mul_(beta2t).add_(1.0 - beta2t, update.mean(dim=-1))
                    exp_avg_sq_col.mul_(beta2t).add_(1.0 - beta2t, update.mean(dim=-2))

                    # Approximation of exponential moving average of square of gradient
                    update = self._approx_sq_grad(exp_avg_sq_row, exp_avg_sq_col)
                    update.mul_(grad)
                else:
                    exp_avg_sq = state["exp_avg_sq"]

                    exp_avg_sq.mul_(beta2t).add_(1.0 - beta2t, update)
                    update = exp_avg_sq.rsqrt().mul_(grad)

                update.div_((self._rms(update) / group["clip_threshold"]).clamp_(min=1.0))
                update.mul_(group["lr"])

                if use_first_moment:
                    exp_avg = state["exp_avg"]
                    exp_avg.mul_(group["beta1"]).add_(1 - group["beta1"], update)
                    update = exp_avg

                if group["weight_decay"] != 0:
                    p_data_fp32.add_(-group["weight_decay"] * group["lr"], p_data_fp32)

                p_data_fp32.add_(-update)

                if p.data.dtype in {torch.float16, torch.bfloat16}:
                    p.data.copy_(p_data_fp32)

        return loss



import logging
import os,sys
import csv
from typing import Dict, List
import numpy as np
from seqeval.metrics import accuracy_score,classification_report,f1_score,precision_score,recall_score
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from log import logging
#logging.root.setLevel(logging.INFO)

logger = logging.getLogger(__name__)

class nerEvaluator:
    """
    This evaluator can be used with the CrossEncoder class.
    It is designed for CrossEncoders with 2 or more outputs. It measure the
    accuracy of the predict class vs. the gold labels.
    """
    def __init__(self, label2id, seq_in: List[str], seq_out: List[str], name: str='', write_csv: bool = True):
        self.label2id=label2id
        self.seq_in = seq_in

        if type(seq_in[0])==list:
            self.input_is_list=True
        else:
            self.input_is_list=False
        if type(seq_in[0])==str:
            self.input_is_string=True
        else:
            self.input_is_string=False

        self.seq_out = seq_out
        self.name = name
        #传进去的seq_in是一个string
        self.csv_file = "nerEvaluator" + ("_" + name if name else '') + "_results.csv"
        self.csv_headers = ["epoch", "steps", "Accuracy", "Precision", "Recall", "F1"]
        self.write_csv = write_csv

        logger.info('Total evaluate nums : {}'.format(len(self.seq_in)))
        logger.info("input is string : {}, input is list : {}".format(self.input_is_string,self.input_is_list))
        logger.info("seq in and out like : \n"+str(self.seq_in[0])+"\t"+str(self.seq_out[0]))
        logger.info("In this evaluator, slot contains "+"("+" ".join(list(self.label2id.keys()))+")")
        #self.slotmetric={key:{} for key in self.label2id.keys()}#记录每一个key的precision,recall and f1 score

    @classmethod
    def from_input_string(cls, label2id, seq_in: List[str], seq_out: List[str], **kwargs):
        return cls(label2id,seq_in,seq_out,**kwargs)


    def __call__(self, model, output_path: str = None, epoch: int = -1, steps: int = -1) -> float:
        if epoch != -1:
            if steps == -1:
                out_txt = " after epoch {}:".format(epoch)
            else:
                out_txt = " in epoch {} after {} steps:".format(epoch, steps)
        else:
            out_txt = ":"

        logger.info("nerEvaluator: Evaluating the model on " + self.name + " dataset" + out_txt)
        pred_tags = model.predict(self.seq_in, input_is_string=self.input_is_string, input_is_list=self.input_is_list, show_progress_bar=False)##(num_eval_examples,seq_len)
        #每一个的长度是不一样的
        golden_tags = []
        for i,each_predict_tag in enumerate(pred_tags):
            if type(self.seq_out[i])==str:
                each_golden_label=self.seq_out[i].strip().split(' ')
            else:
                each_golden_label=self.seq_out[i]
            #print(each_predict_tag,each_golden_label,len(each_predict_tag),len(each_golden_label))
            assert len(each_predict_tag)==len(each_golden_label)
            golden_tags.append(each_golden_label)

        result=classification_report(y_pred=pred_tags, y_true=golden_tags, digits=4)
        f1=f1_score(y_pred=pred_tags, y_true=golden_tags)
        acc=accuracy_score(y_pred=pred_tags, y_true=golden_tags)
        precision=precision_score(y_pred=pred_tags, y_true=golden_tags)
        recall=recall_score(y_pred=pred_tags, y_true=golden_tags)

        print(result)
        assert len(self.seq_in)==len(pred_tags)
        with open(os.path.join(output_path,"predict_tags.txt"),"w",encoding="utf-8") as f:
            for text,predicts in zip(self.seq_in,pred_tags):
                if type(text)==list:
                    text=' '.join(text)
                f.write(text+"\t"+" ".join(predicts)+"\n")

        if output_path is not None and self.write_csv:
            csv_path = os.path.join(output_path, self.csv_file)
            output_file_exists = os.path.isfile(csv_path)
            with open(csv_path, mode="a" if output_file_exists else 'w', encoding="utf-8") as f:
                writer = csv.writer(f)
                if not output_file_exists:
                    writer.writerow(self.csv_headers)

                writer.writerow([epoch, steps, acc, precision, recall, f1])

        return f1


import json
import numpy as np
import os,sys
from typing import Dict, Sequence, Type, Callable, List, Optional
from tqdm import tqdm

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm, trange
from tensorboardX import SummaryWriter
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
#print(sys.path)
from heads import ClsHead
from log import logging

import readers.cls as single_cls
import readers.paircls as pair_cls
from modules.utils import get_optimizer,get_scheduler
from .utils import batch_to_device

logger=logging.getLogger(__name__)

class cls():
    '''
    分类任务上，除了readers中区分单句子还是双句子，
    heads中是不区分的，
    在tasks中只需在fit时指明is_pairs即可
    '''
    def __init__(self,model_path,
                label2id=None,
                max_seq_length:int = 64,
                device:str = None,
                state_dict=None,
                is_finetune=False,
                tensorboard_logdir = None):

        if is_finetune:
            logger.info("Model has been finetuned, so the label2id must be saved in {}".format(os.path.join(model_path,'label2id.json')))
            with open(os.path.join(model_path,'label2id.json')) as f:
                label2id=json.load(f)
        else:
            assert label2id is not None

        self.label2id=label2id
        self.num_labels=len(label2id)
        self.max_seq_lenth=max_seq_length
        logger.info("The label2id is\n {}".format(json.dumps(self.label2id,ensure_ascii=False)))

        self.model=ClsHead(model_path=model_path,
                            num_labels=self.num_labels,
                            state_dict=state_dict,
                            is_finetune=is_finetune)
        if tensorboard_logdir!=None:
            self.tensorboard_writer=SummaryWriter(tensorboard_logdir)
        else:
            self.tensorboard_writer=None

        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"
            logging.info("Use pytorch device: {}".format(device))
        self._target_device = torch.device(device)
        self.model.to(self._target_device)

    def smart_batching_collate_of_single(self,batch):
        features,labels=single_cls.convert_examples_to_features(examples=batch,tokenizer=self.model.tokenizer,max_seq_len=self.max_seq_lenth)
        return features,labels
    
    def smart_batching_collate_of_pair(self,batch):
        features,labels=pair_cls.convert_examples_to_features(examples=batch,tokenizer=self.model.tokenizer,max_seq_len=self.max_seq_lenth)
        return features,labels
    
    def fit(self,
            is_pairs,
            train_dataloader,
            evaluator = None,
            epochs: int = 1,
            loss_fct = None,
            scheduler: str = 'WarmupLinear',
            warmup_proportion: float = 0.1,
            optimizer_type = 'AdamW',
            optimizer_params: Dict[str, object] = {'lr': 2e-5},
            weight_decay: float = 0.01,
            evaluation_steps = None,
            output_path: str = None,
            save_best_model: bool = True,
            max_grad_norm: float = 1,
            use_amp: bool = False,
            callback: Callable[[float, int, int], None] = None,
            show_progress_bar: bool = True,
            early_stop_patience = 10,
            print_loss_step: Optional[int] = None,
            output_all_encoded_layers: bool = False
            ):
        if not os.path.exists(os.path.join(output_path,"label2id.json")):
            with open(os.path.join(output_path,"label2id.json"),'w') as f:
                json.dump(self.label2id,fp=f)#及时保存label2id
            logger.info("label2id has been saved in {}".format(os.path.join(output_path,"label2id.json")))
        #适用于单句子分类和双句子分类任务
        if is_pairs==False:
            logger.info("当前是单句子分类任务")
            train_dataloader.collate_fn=self.smart_batching_collate_of_single
        else:
            logger.info("当前是双句子分类任务")
            train_dataloader.collate_fn=self.smart_batching_collate_of_pair

        if print_loss_step==None:
            print_loss_step=len(train_dataloader)//5
        if evaluator is not None and evaluation_steps==None:
            evaluation_steps=len(train_dataloader)//2
        #一个epoch下打印5次loss，评估2次
        logger.info("一个epoch 下，每隔{}个step会输出一次loss，每隔{}个step会评估一次模型".format(print_loss_step,evaluation_steps))
        if use_amp:
            from torch.cuda.amp import autocast
            scaler = torch.cuda.amp.GradScaler()

        self.model.to(self._target_device)#map_location='cpu'

        if output_path is not None:
            os.makedirs(output_path, exist_ok=True)

        self.best_score = -9999999
        num_train_steps = int(len(train_dataloader) * epochs)
        warmup_steps = num_train_steps*warmup_proportion

        optimizer = get_optimizer(model=self.model,optimizer_type=optimizer_type,weight_decay=weight_decay,optimizer_params=optimizer_params)
        scheduler = get_scheduler(optimizer, scheduler=scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)
        if loss_fct is None:
            loss_fct = nn.BCEWithLogitsLoss() if self.num_labels == 1 else nn.CrossEntropyLoss()

        global_step=0
        skip_scheduler = False
        patience = early_stop_patience        
        for epoch in trange(epochs, desc="Epoch", disable=not show_progress_bar):
            training_steps=0
            training_loss=0.0

            self.model.zero_grad()
            self.model.train()

            for train_step,(features,labels) in tqdm(enumerate(train_dataloader)):
                features=batch_to_device(features,target_device=self._target_device)
                labels=labels.to(self._target_device)
                features['output_all_encoded_layers']=output_all_encoded_layers
                #print(features.keys(),features["input_ids"].size())
                if use_amp:
                    with autocast():
                        logits=self.model(**features)
                        loss_value = loss_fct(logits, labels)

                    scale_before_step = scaler.get_scale()
                    scaler.scale(loss_value).backward()
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                    scaler.step(optimizer)
                    scaler.update()

                    skip_scheduler = scaler.get_scale() != scale_before_step
                else:
                    logits=self.model(**features)
                    loss_value=loss_fct(logits,labels)#CrossEntropyLoss
                    training_loss+=loss_value.item()
                    if print_loss_step!=None and train_step>0 and train_step%print_loss_step == 0:
                        training_loss/=print_loss_step
                        logging.info("Epoch : {}, train_step : {}/{}, loss_value : {} ".format(epoch,train_step*(epoch+1),num_train_steps,training_loss))
                        training_loss=0.0
                    if self.tensorboard_writer is not None:
                        self.tensorboard_writer.add_scalar(f"train_loss",loss_value.item(),global_step=global_step)
                    loss_value.backward()
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                    optimizer.step()

                optimizer.zero_grad()
                if not skip_scheduler:
                    scheduler.step()

                training_steps += 1
                global_step += 1

                if evaluation_steps>0 and evaluator is not None and (train_step) % evaluation_steps == 0 :
                    eval_score=self._eval_during_training(evaluator, output_path, epoch, training_steps, callback)
                    if self.tensorboard_writer is not None:
                        self.tensorboard_writer.add_scalar(f"eval_score",float(eval_score),global_step=global_step)
                    if eval_score > self.best_score:
                        if save_best_model:
                            self.model.save(output_path=output_path)
                            logging.info(("In epoch {}, training_step {}, the eval score is {}, previous eval score is {}, model has been saved in {}".format(epoch,train_step*(epoch+1),eval_score,self.best_score,output_path)))
                            self.best_score=eval_score
                    else:
                        patience-=1
                        logging.info(f"No improvement over previous best eval score ({eval_score:.6f} vs {self.best_score:.6f}), patience = {patience}")
                        if patience==0:
                            logging.info("Run our of patience, early stop!")
                            return
                
                    self.model.zero_grad()
                    self.model.train()

    def _eval_during_training(self, evaluator, output_path,  epoch, steps, callback):
        if evaluator is not None:
            score_and_auc = evaluator(self, label2id=self.label2id, output_path=output_path, epoch=epoch, steps=steps)
            if callback is not None:
                callback(score_and_auc, epoch, steps)
            return score_and_auc
        return None

    def predict(self,
                is_pairs,
                dataloader,
                batch_size:int=32,
                num_workers: int = 0,
                convert_to_numpy: bool = True,
                convert_to_tensor: bool = False,
                output_all_encoded_layers = False,
                show_progress_bar=False):
        if isinstance(dataloader,list):
            #传进来的dataloader是一个List
            dataloader=DataLoader(dataloader,batch_size=batch_size,num_workers=num_workers,shuffle=False)
        if is_pairs==False:
            logger.info('当前是单句子分类任务预测')
            dataloader.collate_fn=self.smart_batching_collate_of_single
        else:
            logger.info("当前是双句子分类任务预测")
            dataloader.collate_fn=self.smart_batching_collate_of_pair
        
        predictions=[]
        self.model.eval()#不仅仅是bert_model
        self.model.to(self._target_device)
        with torch.no_grad():
            for _ in trange(1, desc="Evaluating", disable=not show_progress_bar):
                for features,_ in tqdm(dataloader):
                    features['output_all_encoded_layers']=output_all_encoded_layers
                    features=batch_to_device(features,self._target_device)
                    logits=self.model(**features)
                    probs=torch.nn.functional.softmax(logits,dim=1)#(batch_size,num_labels)
                    predictions.extend(probs)
        if convert_to_tensor:
            predictions=torch.stack(predictions)#(num_eval_examples,num_classes)
        
        elif convert_to_numpy:
            predictions=np.asarray([predict.cpu().detach().numpy() for predict in predictions])

        return predictions#(num_eval_examples,num_classes)


# coding=utf-8
# Copyright 2018 The Google AI Language Team Authors and The HuggingFace Inc. team.
# Copyright (c) 2018, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" BERT model configuration """

import copy
import json
import os,sys
from typing import Any, Dict, Tuple

WEIGHTS_NAME = "pytorch_model.bin"
CONFIG_NAME = "config.json"
#sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from .log import logging
logger = logging.getLogger(__name__)

class PretrainedConfig(object):
    model_type: str = ""

    def __init__(self, **kwargs):
        # Attributes with defaults
        self.return_dict = kwargs.pop("return_dict", False)
        self.output_hidden_states = kwargs.pop("output_hidden_states", False)
        self.output_attentions = kwargs.pop("output_attentions", False)
        self.use_cache = kwargs.pop("use_cache", True)  # Not used by all models
        self.torchscript = kwargs.pop("torchscript", False)  # Only used by PyTorch models
        self.use_bfloat16 = kwargs.pop("use_bfloat16", False)
        self.pruned_heads = kwargs.pop("pruned_heads", {})
        self.tie_word_embeddings = kwargs.pop(
            "tie_word_embeddings", True
        )  # Whether input and output word embeddings should be tied for all MLM, LM and Seq2Seq models.

        # Is decoder is used in encoder-decoder models to differentiate encoder from decoder
        self.is_encoder_decoder = kwargs.pop("is_encoder_decoder", False)
        self.is_decoder = kwargs.pop("is_decoder", False)
        self.add_cross_attention = kwargs.pop("add_cross_attention", False)
        self.tie_encoder_decoder = kwargs.pop("tie_encoder_decoder", False)

        # Parameters for sequence generation
        self.max_length = kwargs.pop("max_length", 20)
        self.min_length = kwargs.pop("min_length", 0)
        self.do_sample = kwargs.pop("do_sample", False)
        self.early_stopping = kwargs.pop("early_stopping", False)
        self.num_beams = kwargs.pop("num_beams", 1)
        self.temperature = kwargs.pop("temperature", 1.0)
        self.top_k = kwargs.pop("top_k", 50)
        self.top_p = kwargs.pop("top_p", 1.0)
        self.repetition_penalty = kwargs.pop("repetition_penalty", 1.0)
        self.length_penalty = kwargs.pop("length_penalty", 1.0)
        self.no_repeat_ngram_size = kwargs.pop("no_repeat_ngram_size", 0)
        self.bad_words_ids = kwargs.pop("bad_words_ids", None)
        self.num_return_sequences = kwargs.pop("num_return_sequences", 1)
        self.chunk_size_feed_forward = kwargs.pop("chunk_size_feed_forward", 0)

        # Fine-tuning task arguments
        self.architectures = kwargs.pop("architectures", None)
        self.finetuning_task = kwargs.pop("finetuning_task", None)
        self.id2label = kwargs.pop("id2label", None)
        self.label2id = kwargs.pop("label2id", None)
        if self.id2label is not None:
            kwargs.pop("num_labels", None)
            self.id2label = dict((int(key), value) for key, value in self.id2label.items())
            # Keys are always strings in JSON so convert ids to int here.
        else:
            self.num_labels = kwargs.pop("num_labels", 2)

        # Tokenizer arguments TODO: eventually tokenizer and models should share the same config
        self.tokenizer_class = kwargs.pop("tokenizer_class", None)
        self.prefix = kwargs.pop("prefix", None)
        self.bos_token_id = kwargs.pop("bos_token_id", None)
        self.pad_token_id = kwargs.pop("pad_token_id", None)
        self.eos_token_id = kwargs.pop("eos_token_id", None)
        self.sep_token_id = kwargs.pop("sep_token_id", None)

        self.decoder_start_token_id = kwargs.pop("decoder_start_token_id", None)

        # task specific arguments
        self.task_specific_params = kwargs.pop("task_specific_params", None)

        # TPU arguments
        self.xla_device = kwargs.pop("xla_device", None)

        # Name or path to the pretrained checkpoint
        self._name_or_path = str(kwargs.pop("name_or_path", ""))

        # Additional attributes without default values
        for key, value in kwargs.items():
            try:
                setattr(self, key, value)
            except AttributeError as err:
                logger.error("Can't set {} with value {} for {}".format(key, value, self))
                raise err

    @property
    def name_or_path(self) -> str:
        return self._name_or_path

    @name_or_path.setter
    def name_or_path(self, value):
        self._name_or_path = str(value)  # Make sure that name_or_path is a string (for JSON encoding)

    @property
    def use_return_dict(self) -> bool:
        """
        :obj:`bool`: Whether or not return :class:`~transformers.file_utils.ModelOutput` instead of tuples.
        """
        # If torchscript is set, force `return_dict=False` to avoid jit errors
        return self.return_dict and not self.torchscript

    @property
    def num_labels(self) -> int:
        """
        :obj:`int`: The number of labels for classification models.
        """
        return len(self.id2label)

    @num_labels.setter
    def num_labels(self, num_labels: int):
        self.id2label = {i: "LABEL_{}".format(i) for i in range(num_labels)}
        self.label2id = dict(zip(self.id2label.values(), self.id2label.keys()))

    def save_pretrained(self, save_directory: str):
        """
        Save a configuration object to the directory ``save_directory``, so that it can be re-loaded using the
        :func:`~transformers.PretrainedConfig.from_pretrained` class method.

        Args:
            save_directory (:obj:`str`):
                Directory where the configuration JSON file will be saved (will be created if it does not exist).
        """
        if os.path.isfile(save_directory):
            raise AssertionError("Provided path ({}) should be a directory, not a file".format(save_directory))
        os.makedirs(save_directory, exist_ok=True)
        # If we save using the predefined names, we can load using `from_pretrained`
        output_config_file = os.path.join(save_directory, CONFIG_NAME)

        self.to_json_file(output_config_file, use_diff=True)
        logger.info("Configuration saved in {}".format(output_config_file))

    @classmethod
    def from_pretrained(cls, pretrained_model_name_or_path: str, **kwargs):
        config_dict, kwargs = cls.get_config_dict(pretrained_model_name_or_path, **kwargs)
        return cls.from_dict(config_dict, **kwargs)

    @classmethod
    def get_config_dict(cls, pretrained_model_name_or_path: str, **kwargs) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        assert os.path.isdir(pretrained_model_name_or_path)
        config_file = os.path.join(pretrained_model_name_or_path, CONFIG_NAME)
        try:
            config_dict = cls._dict_from_json_file(config_file)
        except EnvironmentError:
            msg = (
                f"Can't load config for '{pretrained_model_name_or_path}'. Make sure that:\n\n"
                f"- '{pretrained_model_name_or_path}' is a correct model identifier listed on 'https://huggingface.co/models'\n\n"
                f"- or '{pretrained_model_name_or_path}' is the correct path to a directory containing a {CONFIG_NAME} file\n\n"
            )
            raise EnvironmentError(msg)
        except json.JSONDecodeError:
            msg = (
                "Couldn't reach server at '{}' to download configuration file or "
                "configuration file is not a valid JSON file. "
                "Please check network or file content here: {}.".format(config_file, config_file)
            )
            raise EnvironmentError(msg)
        logger.info("loading configuration file {}".format(config_file))
        return config_dict, kwargs

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any], **kwargs):
        return_unused_kwargs = kwargs.pop("return_unused_kwargs", False)

        config = cls(**config_dict)

        if hasattr(config, "pruned_heads"):
            config.pruned_heads = dict((int(key), value) for key, value in config.pruned_heads.items())

        # Update config with kwargs if needed
        to_remove = []
        for key, value in kwargs.items():
            if hasattr(config, key):
                setattr(config, key, value)
                to_remove.append(key)
        for key in to_remove:
            kwargs.pop(key, None)

        #logger.info("Model config %s", str(config))
        if return_unused_kwargs:
            return config, kwargs
        else:
            return config

    @classmethod
    def from_json_file(cls, json_file: str) -> "PretrainedConfig":
        """
        Instantiates a :class:`~transformers.PretrainedConfig` from the path to a JSON file of parameters.

        Args:
            json_file (:obj:`str`):
                Path to the JSON file containing the parameters.

        Returns:
            :class:`PretrainedConfig`: The configuration object instantiated from that JSON file.

        """
        config_dict = cls._dict_from_json_file(json_file)
        return cls(**config_dict)

    @classmethod
    def _dict_from_json_file(cls, json_file: str):
        with open(json_file, "r", encoding="utf-8") as reader:
            text = reader.read()
        return json.loads(text)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return "{} {}".format(self.__class__.__name__, self.to_json_string())

    def to_diff_dict(self) -> Dict[str, Any]:
        """
        Removes all attributes from config which correspond to the default
        config attributes for better readability and serializes to a Python
        dictionary.

        Returns:
            :obj:`Dict[str, Any]`: Dictionary of all the attributes that make up this configuration instance,
        """
        config_dict = self.to_dict()

        # get the default config dict
        default_config_dict = PretrainedConfig().to_dict()

        serializable_config_dict = {}

        # only serialize values that differ from the default config
        for key, value in config_dict.items():
            if key not in default_config_dict or value != default_config_dict[key]:
                serializable_config_dict[key] = value

        return serializable_config_dict

    def to_dict(self) -> Dict[str, Any]:
        """
        Serializes this instance to a Python dictionary.

        Returns:
            :obj:`Dict[str, Any]`: Dictionary of all the attributes that make up this configuration instance.
        """
        output = copy.deepcopy(self.__dict__)
        if hasattr(self.__class__, "model_type"):
            output["model_type"] = self.__class__.model_type
        return output

    def to_json_string(self, use_diff: bool = True) -> str:
        """
        Serializes this instance to a JSON string.

        Args:
            use_diff (:obj:`bool`, `optional`, defaults to :obj:`True`):
                If set to ``True``, only the difference between the config instance and the default
                ``PretrainedConfig()`` is serialized to JSON string.

        Returns:
            :obj:`str`: String containing all the attributes that make up this configuration instance in JSON format.
        """
        if use_diff is True:
            config_dict = self.to_diff_dict()
        else:
            config_dict = self.to_dict()
        return json.dumps(config_dict, indent=2, sort_keys=True) + "\n"

    def to_json_file(self, json_file_path: str, use_diff: bool = True):
        """
        Save this instance to a JSON file.

        Args:
            json_file_path (:obj:`str`):
                Path to the JSON file in which this configuration instance's parameters will be saved.
            use_diff (:obj:`bool`, `optional`, defaults to :obj:`True`):
                If set to ``True``, only the difference between the config instance and the default
                ``PretrainedConfig()`` is serialized to JSON file.
        """
        with open(json_file_path, "w", encoding="utf-8") as writer:
            writer.write(self.to_json_string(use_diff=use_diff))

    def update(self, config_dict: Dict[str, Any]):
        """
        Updates attributes of this class with attributes from ``config_dict``.

        Args:
            config_dict (:obj:`Dict[str, Any]`): Dictionary of attributes that shall be updated for this class.
        """
        for key, value in config_dict.items():
            setattr(self, key, value)

class BertConfig(PretrainedConfig):
    r"""
    This is the configuration class to store the configuration of a :class:`~transformers.BertModel` or a
    :class:`~transformers.TFBertModel`. It is used to instantiate a BERT model according to the specified
    arguments, defining the model architecture. Instantiating a configuration with the defaults will yield a similar
    configuration to that of the BERT `bert-base-uncased <https://huggingface.co/bert-base-uncased>`__ architecture.

    Configuration objects inherit from  :class:`~transformers.PretrainedConfig` and can be used
    to control the model outputs. Read the documentation from  :class:`~transformers.PretrainedConfig`
    for more information.


    Args:
        vocab_size (:obj:`int`, `optional`, defaults to 30522):
            Vocabulary size of the BERT model. Defines the number of different tokens that can be represented by the
            :obj:`inputs_ids` passed when calling :class:`~transformers.BertModel` or
            :class:`~transformers.TFBertModel`.
        hidden_size (:obj:`int`, `optional`, defaults to 768):
            Dimensionality of the encoder layers and the pooler layer.
        num_hidden_layers (:obj:`int`, `optional`, defaults to 12):
            Number of hidden layers in the Transformer encoder.
        num_attention_heads (:obj:`int`, `optional`, defaults to 12):
            Number of attention heads for each attention layer in the Transformer encoder.
        intermediate_size (:obj:`int`, `optional`, defaults to 3072):
            Dimensionality of the "intermediate" (often named feed-forward) layer in the Transformer encoder.
        hidden_act (:obj:`str` or :obj:`Callable`, `optional`, defaults to :obj:`"gelu"`):
            The non-linear activation function (function or string) in the encoder and pooler.
            If string, :obj:`"gelu"`, :obj:`"relu"`, :obj:`"swish"` and :obj:`"gelu_new"` are supported.
        hidden_dropout_prob (:obj:`float`, `optional`, defaults to 0.1):
            The dropout probabilitiy for all fully connected layers in the embeddings, encoder, and pooler.
        attention_probs_dropout_prob (:obj:`float`, `optional`, defaults to 0.1):
            The dropout ratio for the attention probabilities.
        max_position_embeddings (:obj:`int`, `optional`, defaults to 512):
            The maximum sequence length that this model might ever be used with.
            Typically set this to something large just in case (e.g., 512 or 1024 or 2048).
        type_vocab_size (:obj:`int`, `optional`, defaults to 2):
            The vocabulary size of the :obj:`token_type_ids` passed when calling :class:`~transformers.BertModel` or
            :class:`~transformers.TFBertModel`.
        initializer_range (:obj:`float`, `optional`, defaults to 0.02):
            The standard deviation of the truncated_normal_initializer for initializing all weight matrices.
        layer_norm_eps (:obj:`float`, `optional`, defaults to 1e-12):
            The epsilon used by the layer normalization layers.
        gradient_checkpointing (:obj:`bool`, `optional`, defaults to :obj:`False`):
            If True, use gradient checkpointing to save memory at the expense of slower backward pass.

    Examples::

        >>> from transformers import BertModel, BertConfig

        >>> # Initializing a BERT bert-base-uncased style configuration
        >>> configuration = BertConfig()

        >>> # Initializing a model from the bert-base-uncased style configuration
        >>> model = BertModel(configuration)

        >>> # Accessing the model configuration
        >>> configuration = model.config
    """
    model_type = "bert"

    def __init__(
        self,
        vocab_size=30522,
        hidden_size=768,
        num_hidden_layers=12,
        num_attention_heads=12,
        intermediate_size=3072,
        hidden_act="gelu",
        hidden_dropout_prob=0.1,
        attention_probs_dropout_prob=0.1,
        max_position_embeddings=512,
        type_vocab_size=2,
        initializer_range=0.02,
        layer_norm_eps=1e-12,
        pad_token_id=0,
        gradient_checkpointing=False,
        **kwargs
    ):
        super().__init__(pad_token_id=pad_token_id, **kwargs)

        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.num_hidden_layers = num_hidden_layers
        self.num_attention_heads = num_attention_heads
        self.hidden_act = hidden_act
        self.intermediate_size = intermediate_size
        self.hidden_dropout_prob = hidden_dropout_prob
        self.attention_probs_dropout_prob = attention_probs_dropout_prob
        self.max_position_embeddings = max_position_embeddings
        self.type_vocab_size = type_vocab_size
        self.initializer_range = initializer_range
        self.layer_norm_eps = layer_norm_eps
        self.gradient_checkpointing = gradient_checkpointing



from typing import Text, Union, List
import numpy as np
import random,sys,os
import torch

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from log import logging
logger=logging.getLogger(__name__)

from .base_reader import convert_to_tensor

SPECIAL_TOKEN_NUMS=2#[CLS] and [SEP]

class InputExample:
    """
    Structure for one input example with texts, the label and a unique id
    """
    def __init__(self, guid = '', text = None,  label = 0):
        """
        单句子分类
        text是string类型的句子
        """
        self.guid = guid
        self.text = text
        self.label = label


    def __str__(self):
        return "<InputExample> label: {}, text: {}".format(str(self.label), self.text)

class InputFeatures:
    def __init__(self, input_ids,token_type_ids,attention_mask):
        '''
        单句子分类，token_type_ids应该是全部值为0的list
        '''
        self.input_ids=input_ids
        self.token_type_ids=token_type_ids
        self.attention_mask=attention_mask
    def __str__(self) -> str:
        return "<InputFeature> input_ids: {}\n token_type_ids: {}\n attention_mask: {}\n".format(' '.join([str(x) for x in self.input_ids]),
                                                                                            ' '.join([str(x) for x in self.token_type_ids]),
                                                                                            ' '.join([str(x) for x in self.attention_mask]))


def convert_examples_to_features(examples, tokenizer, max_seq_len):
    if isinstance(examples[0],str):
        #说明此时examples中的每一个元素是一个句子，还没有转为InputExample
        examples=[InputExample(text=text) for text in examples]
        
    features = []
    labels=[]
    assert type(examples[0].text)==str#单句子分类
    max_len_this_batch=0
    for example in examples:
        length=len(tokenizer.tokenize(example.text))
        length+=SPECIAL_TOKEN_NUMS
        if length>max_len_this_batch:
            max_len_this_batch=length

    max_seq_len=min(max_seq_len,max_len_this_batch)
    for example_index, example in enumerate(examples):
        tokens = []
        token_type_ids = []
        tokens.append("[CLS]")
        token_type_ids.append(0)
        
        tokens_text_a = tokenizer.tokenize(example.text)
        tokens += tokens_text_a
        token_type_ids += [0] * len(tokens_text_a)
        tokens.append('[SEP]')
        token_type_ids.append(0)

        if len(tokens) >= max_seq_len:
            tokens = tokens[:max_seq_len]
            token_type_ids = token_type_ids[:max_seq_len]

        seq_len = len(tokens)
        pad_len = max_seq_len - seq_len
        # print(tokens)
        input_ids = tokenizer.convert_tokens_to_ids(tokens)
        input_ids += [0] * pad_len
        attention_mask = [1] * seq_len + [0] * pad_len
        token_type_ids += [0] * pad_len
        assert len(input_ids) == len(token_type_ids) == len(attention_mask) == max_seq_len
        labels.append(example.label)

        features.append(InputFeatures(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask))
        # if example_index < 5:
        #     logging.info("*** Example ***")
        #     logging.info("example_index: %s" % (example_index))
        #     logging.info("tokens: %s" % " ".join(tokens))
        #     logging.info("input_ids: %s" % " ".join([str(x) for x in input_ids]))
        #     logging.info("attention_mask: %s" % " ".join([str(x) for x in attention_mask]))
        #     logging.info("token_type_ids: %s" % " ".join([str(x) for x in token_type_ids]))
        #     logging.info("label_ids : %s" % " ".join([str(example.label)]))
    
    features=convert_to_tensor(features=features)
    labels=torch.LongTensor(labels)
    return features,labels

def getExamplesFromFile(file_path,label2id,mode="train"):
    with open(file_path, encoding='utf-8') as f:
        lines = f.readlines()
    examples=[]
    for line in lines:
        line_split = line.strip().split('\t')
        sentence,label = line_split
        assert type(sentence)==str

        examples.append(InputExample(text=sentence,label=label2id[label]))

    logger.info("*****************************Logging some {} examples*****************************".format(mode))
    logger.info("Total {} nums is : {}".format(mode,len(examples)))
    for _ in range(5):
        i=random.randint(0,len(examples)-1)
        logger.info('\n'+examples[i].text+"\t"+str(examples[i].label))
    return examples

def getExamplesFromData(sentences,labels,label2id,mode="train",return_max_len=False):
    examples=[]
    max_seq_len=0
    for sentence,label in zip(sentences,labels):
        if len(sentence)>max_seq_len:
            max_seq_len=len(sentence)
        examples.append(InputExample(text=sentence,label=label2id[str(label)]))

    logger.info("*****************************Logging some {} examples*****************************".format(mode))
    logger.info("Total {} nums is : {}".format(mode,len(examples)))
    for _ in range(5):
        i=random.randint(0,len(examples)-1)
        logger.info('\n'+examples[i].text+"\t"+str(examples[i].label))

    if return_max_len:
        return examples,max_seq_len
    else:
        return examples




import json
from logging import log
import numpy as np
import os,sys
from typing import Dict, Sequence, Text, Type, Callable, List, Optional
from tqdm import tqdm

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm, trange
from tensorboardX import SummaryWriter
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
#print(sys.path)
from heads import NerHead
from log import logging

from readers.ner import convert_examples_to_features,InputExample
from modules.utils import get_optimizer,get_scheduler
from .utils import batch_to_device,eval_during_training

logger=logging.getLogger(__name__)

class Ner():
    def __init__(self,model_path,
                label2id,
                max_seq_length:int = 64,
                device:str = None,
                use_bilstm=False,
                use_crf=False,
                state_dict=None,
                is_finetune=False,
                tensorboard_logdir = None,
                batch_first=True):

        if label2id is None:
            assert os.path.exists(os.path.join(model_path,'label2id.json'))
            with open(os.path.join(model_path,'label2id.json')) as f:
                label2id=json.load(f)
            logger.info("Load label2id from {}".format(os.path.join(model_path,'label2id.json')))
        else:
            if not os.path.exists(os.path.join(model_path,"label2id.json")):
                with open(os.path.join(model_path,"label2id.json"),'w') as f:
                    json.dump(label2id,fp=f)#及时保存label2id
            logger.info('label2id has been saved in {}'.format(os.path.join(model_path,'label2id.json')))


        self.label2id=label2id
        self.use_crf=use_crf
        self.use_bilstm=use_bilstm
        self.batch_first=batch_first
        assert self.label2id["[PAD]"]==0#将[CLS],[SEP],[PAD]这三个特殊字符全部都看作[PAD]处理，也就是将[PAD]作为一个标签
        self.id2label={id_:label for label,id_ in self.label2id.items()}
        self.num_labels=len(label2id)
        self.max_seq_lenth=max_seq_length
        logger.info("The label2id is\n {}".format(json.dumps(self.label2id,ensure_ascii=False)))

        self.model=NerHead(model_path=model_path,
                            num_labels=self.num_labels,
                            state_dict=state_dict,
                            is_finetune=is_finetune,
                            use_bilstm=use_bilstm,
                            use_crf=use_crf)

        if tensorboard_logdir!=None:
            self.tensorboard_writer=SummaryWriter(tensorboard_logdir)
        else:
            self.tensorboard_writer=None

        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"
            logging.info("Use pytorch device: {}".format(device))
        self._target_device = torch.device(device)
        self.model.to(self._target_device)
        logger.info("Using BiLSTM ? : {}".format(self.use_bilstm))
        logger.info("Using CRF ? : {}".format(self.use_crf))
        
    def smart_batching_collate(self,batch):
        features=convert_examples_to_features(examples=batch,tokenizer=self.model.tokenizer,max_seq_len=self.max_seq_lenth,label2id=self.label2id)
        return features#features.keys()==[input_ids,token_type_ids,attention_mask,label_ids]

    def getLoss(self,logits,labels,attention_mask):
        '''
        last_hidden_state.size()==(batch_size,seq_length,hidden_size) 
        labels.size()==(batch_size,seq_length)
        '''
        if self.use_crf:
            loss=self.model.crfLayer(emissions=logits,tags=labels,mask=attention_mask)
        else:
            mask_pad=attention_mask.view(-1)==1
            logits=logits.view(-1,self.num_labels)[mask_pad]
            labels=labels.view(-1)[mask_pad]
            loss=nn.CrossEntropyLoss(ignore_index=0)(logits,labels)
        # label_one_hot=nn.functional.one_hot(labels,num_classes=self.num_labels).to(labels.device)
        # #print(label_one_hot.size(),logits.size())
        # assert logits.size()==(self.current_batch_size,self.current_seq_len,self.num_labels)==label_one_hot.size()
        # loss=-torch.sum(torch.log_softmax(logits,dim=-1)*label_one_hot)/(self.current_batch_size*self.current_seq_len)
        return loss

    def maskPredictions(self,predictions,attention_mask):
        if type(attention_mask)!=list:
            attention_mask=attention_mask.cpu().tolist()
        masked_result=[]
        for prediction,mask in zip(predictions,attention_mask):
            if 0 not in mask:
                #[1,1,1,1,1,1,1]#此时[SEP]是最后一个
                masked_result.append(prediction[1:-1])
            else:
                #[1,1,1,1,1,1,0]#此时[SEP]是倒数第二个
                end_pos=mask.index(0)
                masked_result.append(prediction[1:end_pos-1])
        return masked_result

    def getPredictions(self,features):
        features=batch_to_device(features,self._target_device)
        attention_mask=features["attention_mask"]#(batch_size,pad_seq_len)
        logits=self.model(**features)#(batch_size,pad_seq_len,num_labels)
        probs=torch.nn.functional.softmax(logits,dim=2)
        if not self.use_crf:
            predicts=torch.argmax(probs,dim=2).cpu().numpy().tolist()
        else:
            predicts=self.model.crfLayer.viterbi_decode(emissions=logits,mask=attention_mask)
        predicts=self.maskPredictions(predictions=predicts,attention_mask=attention_mask)
        #在viterbi_decode返回的已经是不包含有mask的list了
        predicts_tags=[]
        for predict in predicts:
            temp=[]
            for id_ in predict:
                tag=self.id2label[id_]
                temp.append(tag)
            predicts_tags.append(temp)
        return predicts_tags

    def fit(self,
            train_dataloader,
            evaluator = None,
            epochs: int = 1,
            scheduler: str = 'WarmupLinear',
            warmup_proportion: float = 0.1,
            optimizer_type = 'AdamW',
            optimizer_params: Dict[str, object] = {'lr': 2e-5},
            weight_decay: float = 0.01,
            evaluation_steps = None,
            output_path: str = None,
            save_best_model: bool = True,
            max_grad_norm: float = 1,
            use_amp: bool = False,
            callback: Callable[[float, int, int], None] = None,
            show_progress_bar: bool = True,
            early_stop_patience = 10,
            print_loss_step: Optional[int] = None,
            output_all_encoded_layers: bool = False
            ):
        train_dataloader.collate_fn=self.smart_batching_collate

        if print_loss_step==None:
            print_loss_step=len(train_dataloader)//5
        if evaluator is not None and evaluation_steps==None:
            evaluation_steps=len(train_dataloader)//2
        #一个epoch下打印5次loss，评估2次
        logger.info("一个epoch 下，每隔{}个step会输出一次loss，每隔{}个step会评估一次模型".format(print_loss_step,evaluation_steps))
        if use_amp:
            from torch.cuda.amp import autocast
            scaler = torch.cuda.amp.GradScaler()

        self.model.to(self._target_device)#map_location='cpu'

        if output_path is not None:
            os.makedirs(output_path, exist_ok=True)

        self.best_score = -9999999
        num_train_steps = int(len(train_dataloader) * epochs)
        warmup_steps = num_train_steps*warmup_proportion

        optimizer = get_optimizer(model=self.model,optimizer_type=optimizer_type,weight_decay=weight_decay,optimizer_params=optimizer_params)
        scheduler = get_scheduler(optimizer, scheduler=scheduler, warmup_steps=warmup_steps, t_total=num_train_steps)

        global_step=0
        skip_scheduler = False
        patience = early_stop_patience        
        for epoch in trange(epochs, desc="Epoch", disable=not show_progress_bar):
            training_steps=0
            training_loss=0.0

            self.model.zero_grad()
            self.model.train()

            for train_step,features_and_labels in tqdm(enumerate(train_dataloader)):
                features_and_labels=batch_to_device(features_and_labels,target_device=self._target_device)
                labels=features_and_labels['label_ids']
                self.current_batch_size,self.current_seq_len=labels.size()
                features_and_labels['output_all_encoded_layers']=output_all_encoded_layers
                #print(features.keys(),features["input_ids"].size())
                if use_amp:
                    with autocast():
                        logits=self.model(**features_and_labels)
                        loss_value = self.getLoss(logits=logits,labels=labels,attention_mask=features_and_labels['attention_mask'])

                    scale_before_step = scaler.get_scale()
                    scaler.scale(loss_value).backward()
                    scaler.unscale_(optimizer)
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                    scaler.step(optimizer)
                    scaler.update()

                    skip_scheduler = scaler.get_scale() != scale_before_step
                else:
                    logits=self.model(**features_and_labels)
                    loss_value = self.getLoss(logits=logits,labels=labels,attention_mask=features_and_labels['attention_mask'])
                    
                    training_loss+=loss_value.item()
                    if print_loss_step!=None and train_step>0 and train_step%print_loss_step == 0:
                        training_loss/=print_loss_step
                        logging.info("Epoch : {}, train_step : {}/{}, loss_value : {} ".format(epoch,train_step*(epoch+1),num_train_steps,training_loss))
                        training_loss=0.0
                    if self.tensorboard_writer is not None:
                        self.tensorboard_writer.add_scalar(f"train_loss",loss_value.item(),global_step=global_step)
                    loss_value.backward()
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_grad_norm)
                    optimizer.step()

                optimizer.zero_grad()
                if not skip_scheduler:
                    scheduler.step()

                training_steps += 1
                global_step += 1

                if evaluation_steps>0 and evaluator is not None and (train_step) % evaluation_steps == 0 :
                    eval_score=eval_during_training(self, evaluator, output_path, epoch, training_steps, callback)
                    if self.tensorboard_writer is not None:
                        self.tensorboard_writer.add_scalar(f"eval_score",float(eval_score),global_step=global_step)
                    if eval_score > self.best_score:
                        if save_best_model:
                            self.model.save(output_path=output_path)
                            logging.info(("In epoch {}, training_step {}, the eval score is {}, previous eval score is {}, model has been saved in {}".format(epoch,train_step*(epoch+1),eval_score,self.best_score,output_path)))
                            self.best_score=eval_score
                    else:
                        patience-=1
                        logging.info(f"No improvement over previous best eval score ({eval_score:.6f} vs {self.best_score:.6f}), patience = {patience}")
                        if patience==0:
                            logging.info("Run our of patience, early stop!")
                            return
                
                    self.model.zero_grad()
                    self.model.train()

    def predict(self,
                dataloader,
                batch_size:int=32,
                num_workers: int = 0,
                output_all_encoded_layers = False,
                show_progress_bar=False,
                input_is_string=True,
                input_is_list=False):
        '''
        nerEvaluator传进来的是nerEvaluator.seq_in，其中每一个元素正是一个句子
        '''
        if type(dataloader)==str:
            dataloader=[dataloader]
        if input_is_string:
            #说明此时传进来的是句子，不是InputExample类型
            dataloader=[InputExample(seq_in=self.model.tokenizer.tokenize(text)) for text in dataloader]
        if input_is_list:
            dataloader=[InputExample(seq_in=seq_in) for seq_in in dataloader]

        dataloader=DataLoader(dataloader,batch_size=batch_size,num_workers=num_workers,shuffle=False)
        dataloader.collate_fn=self.smart_batching_collate
        
        predictions=[]
        self.model.eval()#不仅仅是bert_model
        self.model.to(self._target_device)
        with torch.no_grad():
            for _ in trange(1, desc="Evaluating", disable=not show_progress_bar):
                for features in dataloader:
                    features['output_all_encoded_layers']=output_all_encoded_layers
                    batch_predicts_tags=self.getPredictions(features)
                    predictions.extend(batch_predicts_tags)
        
        return predictions#(num_eval_examples,seq_len)


import os
import csv
from typing import List
import numpy as np
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from log import logging
#logging.root.setLevel(logging.INFO)

logger = logging.getLogger(__name__)

from sklearn.metrics.pairwise import paired_cosine_distances, paired_euclidean_distances, paired_manhattan_distances
from scipy.stats import pearsonr, spearmanr

from enum import Enum

class SimilarityFunction(Enum):
    COSINE = 0
    EUCLIDEAN = 1
    MANHATTAN = 2
    DOT_PRODUCT = 3

class stsEvaluator:
    """
    This evaluator can be used with the CrossEncoder class.
    It is designed for CrossEncoders with 2 or more outputs. It measure the
    accuracy of the predict class vs. the gold labels.
    """
    def __init__(self, sentences1: List[str], sentences2: List[str], scores: List[float], batch_size: int = 16, main_similarity: SimilarityFunction = SimilarityFunction.COSINE, name: str = '', show_progress_bar: bool = True, write_csv: bool = True ,
                    whitening_W = None, whitening_mu = None):
        """
        Constructs an evaluator based for the dataset
        The labels need to indicate the similarity between the sentences.
        :param sentences1:  List with the first sentence in a pair
        :param sentences2: List with the second sentence in a pair
        :param scores: Similarity score between sentences1[i] and sentences2[i]
        :param write_csv: Write results to a CSV file
        """
        self.sentences1 = sentences1
        self.sentences2 = sentences2
        self.scores = scores
        self.write_csv = write_csv

        assert len(self.sentences1) == len(self.sentences2)
        assert len(self.sentences1) == len(self.scores)

        self.main_similarity = main_similarity
        self.name = name

        self.batch_size = batch_size
        self.show_progress_bar = show_progress_bar

        self.csv_file = "similarity_evaluation"+("_"+name if name else '')+"_results.csv"
        self.csv_headers = ["epoch", "steps", "cosine_pearson", "cosine_spearman", "euclidean_pearson", "euclidean_spearman", "manhattan_pearson", "manhattan_spearman", "dot_pearson", "dot_spearman"]

        #whitening
        self.whitening_W=whitening_W
        self.whitening_mu=whitening_mu
        if self.whitening_W is not None:
            logging.info("Using whitening to smooth bert sentence embeddings")
            self.n_components=self.whitening_W.shape[1]

    def __call__(self, model, output_path: str = None, epoch: int = -1, steps: int = -1) -> float:
        if epoch != -1:
            if steps == -1:
                out_txt = " after epoch {}:".format(epoch)
            else:
                out_txt = " in epoch {} after {} steps:".format(epoch, steps)
        else:
            out_txt = ":"

        logging.info("EmbeddingSimilarityEvaluator: Evaluating the model on " + self.name + " dataset" + out_txt)

        embeddings1 = model.encode(self.sentences1, batch_size=self.batch_size, show_progress_bar=self.show_progress_bar, convert_to_numpy=True)
        embeddings2 = model.encode(self.sentences2, batch_size=self.batch_size, show_progress_bar=self.show_progress_bar, convert_to_numpy=True)
        labels = self.scores
        bsz,dim=embeddings1.shape
        #embeddings1.shape==embeddings2.shape==(bsz,768)
        if self.whitening_W is not None:
            #W.shape==(768,n_components), mu.shape==(1,768)
            embeddings1=(embeddings1-self.whitening_mu).dot(self.whitening_W)#(batch_size,n_components)
            embeddings2=(embeddings2-self.whitening_mu).dot(self.whitening_W)#(batch_size,n_components)
            assert embeddings1.shape==embeddings2.shape==(bsz,self.n_components)
        #self.scores就是example.label，这个label通常是1或者0，也可以是浮点数，代表的含义是两个句子的相关度分数
        cosine_scores = 1 - (paired_cosine_distances(embeddings1, embeddings2))#cosine_scores.size()==(batch_size),代表的是两行对应的每一个样本的cosine距离，注意这是距离不是score，所以要1-distance
        manhattan_distances = -paired_manhattan_distances(embeddings1, embeddings2)
        euclidean_distances = -paired_euclidean_distances(embeddings1, embeddings2)
        dot_products = [np.dot(emb1, emb2) for emb1, emb2 in zip(embeddings1, embeddings2)]
        #manhattan_distances and euclidean_distances both is (batch_size,)，come along with scores to compute each metrics

        eval_pearson_cosine, _ = pearsonr(labels, cosine_scores)#second item is p-value
        eval_spearman_cosine, _ = spearmanr(labels, cosine_scores)

        eval_pearson_manhattan, _ = pearsonr(labels, manhattan_distances)
        eval_spearman_manhattan, _ = spearmanr(labels, manhattan_distances)

        eval_pearson_euclidean, _ = pearsonr(labels, euclidean_distances)
        eval_spearman_euclidean, _ = spearmanr(labels, euclidean_distances)

        eval_pearson_dot, _ = pearsonr(labels, dot_products)
        eval_spearman_dot, _ = spearmanr(labels, dot_products)

        logging.info("Cosine-Similarity :\tPearson: {:.4f}\tSpearman: {:.4f}".format(
            eval_pearson_cosine, eval_spearman_cosine))
        logging.info("Manhattan-Distance:\tPearson: {:.4f}\tSpearman: {:.4f}".format(
            eval_pearson_manhattan, eval_spearman_manhattan))
        logging.info("Euclidean-Distance:\tPearson: {:.4f}\tSpearman: {:.4f}".format(
            eval_pearson_euclidean, eval_spearman_euclidean))
        logging.info("Dot-Product-Similarity:\tPearson: {:.4f}\tSpearman: {:.4f}".format(
            eval_pearson_dot, eval_spearman_dot))

        if output_path is not None and self.write_csv:
            csv_path = os.path.join(output_path, self.csv_file)
            output_file_exists = os.path.isfile(csv_path)
            with open(csv_path, newline='', mode="a" if output_file_exists else 'w', encoding="utf-8") as f:
                writer = csv.writer(f)
                if not output_file_exists:
                    writer.writerow(self.csv_headers)

                writer.writerow([epoch, steps, eval_pearson_cosine, eval_spearman_cosine, eval_pearson_euclidean,
                                 eval_spearman_euclidean, eval_pearson_manhattan, eval_spearman_manhattan, eval_pearson_dot, eval_spearman_dot])


        if self.main_similarity == SimilarityFunction.COSINE:
            #default is cosine，也就是说计算各个指标的时候是需要传进去scores的，而目前的scores都是类别1和0代替
            return eval_spearman_cosine
        elif self.main_similarity == SimilarityFunction.EUCLIDEAN:
            return eval_spearman_euclidean
        elif self.main_similarity == SimilarityFunction.MANHATTAN:
            return eval_spearman_manhattan
        elif self.main_similarity == SimilarityFunction.DOT_PRODUCT:
            return eval_spearman_dot
        elif self.main_similarity is None:
            return max(eval_spearman_cosine, eval_spearman_manhattan, eval_spearman_euclidean, eval_spearman_dot)
        else:
            raise ValueError("Unknown main_similarity value")


import torch
from torch import Tensor
from torch import nn
from typing import Union, Tuple, List, Iterable, Dict
import os
import json


class Pooling(nn.Module):
    """
    Pooling层的操作就是根据指定的条件，如(cls,max,mean,first_last,last_2等)对token_embeddings进行相应的pooling操作
    然后拼接每一个pooling操作的result，update到features["sentence_embedding"]输出features。
    这个sentence_embedding的维度显然是768*pooling操作的个数
    self.pooling_output_dimension = (pooling_mode_multiplier * word_embedding_dimension)会将Pooling输出的向量维度返回给
    SentenceTransformer的model.get_pooling_dimension，然后这个值回传给Loss模型，这样就可以知道在loss中知道sentence_embedding的维度了
    loss中的条件是concatenation_sent_rep和difference以及multiplication，然后再计算loss
    也就是说，假如pooling方式选择mean+last-first，那么句子向量维度就是768*2。loss中拼接的条件是rep+difference+multiplication
    那么此时传进给loss的维度是(768*2*2)+(768*2)+(768*2)
    """

    def __init__(self,
                 word_embedding_dimension: int,
                 pooling_mode_cls_token: bool = False,
                 pooling_mode_max_tokens: bool = False,
                 pooling_mode_mean_tokens: bool = True,
                 pooling_mode_mean_sqrt_len_tokens: bool = False,
                 pooling_mode_mean_last_2_tokens: bool = False,
                 pooling_mode_mean_first_last_tokens: bool = False,
                 ):
        super(Pooling, self).__init__()

        self.config_keys = ['word_embedding_dimension', 'pooling_mode_cls_token', 'pooling_mode_mean_tokens',
                            'pooling_mode_max_tokens', 'pooling_mode_mean_sqrt_len_tokens',
                            'pooling_mode_mean_last_2_tokens','pooling_mode_mean_first_last_tokens']

        self.word_embedding_dimension = word_embedding_dimension
        self.pooling_mode_cls_token = pooling_mode_cls_token
        self.pooling_mode_mean_tokens = pooling_mode_mean_tokens
        self.pooling_mode_max_tokens = pooling_mode_max_tokens
        self.pooling_mode_mean_sqrt_len_tokens = pooling_mode_mean_sqrt_len_tokens
        self.pooling_mode_mean_last_2_tokens = pooling_mode_mean_last_2_tokens
        self.pooling_mode_mean_first_last_tokens = pooling_mode_mean_first_last_tokens

        pooling_mode_multiplier = sum([pooling_mode_cls_token, pooling_mode_max_tokens, pooling_mode_mean_tokens,
                                       pooling_mode_mean_sqrt_len_tokens, pooling_mode_mean_last_2_tokens,pooling_mode_mean_first_last_tokens])
        self.pooling_output_dimension = (pooling_mode_multiplier * word_embedding_dimension)

    def forward(self, token_embeddings, cls_token_embeddings, attention_mask, all_layer_embeddings=None, token_weights_sum=None):
        '''
        token_embeddings就是sequence_output，注意sequence_output是包含有cls的
        cls_token_embeddings就是pooled_output
        '''
        #assert features["all_layer_embeddings"][-1].sum() == features["token_embeddings"].sum()

        ## Pooling strategy
        output_vectors = []
        if self.pooling_mode_cls_token:
            output_vectors.append(cls_token_embeddings)
        if self.pooling_mode_max_tokens:
            input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            token_embeddings[input_mask_expanded == 0] = -1e9  # Set padding tokens to large negative value
            max_over_time = torch.max(token_embeddings, 1)[0]
            output_vectors.append(max_over_time)
        if self.pooling_mode_mean_tokens or self.pooling_mode_mean_sqrt_len_tokens:
            input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, 1)

            # If tokens are weighted (by WordWeights layer), feature 'token_weights_sum' will be present
            if token_weights_sum is not None:
                sum_mask = token_weights_sum.unsqueeze(-1).expand(sum_embeddings.size())
            else:
                sum_mask = input_mask_expanded.sum(1)

            sum_mask = torch.clamp(sum_mask, min=1e-9)

            if self.pooling_mode_mean_tokens:
                output_vectors.append(sum_embeddings / sum_mask)
            if self.pooling_mode_mean_sqrt_len_tokens:
                output_vectors.append(sum_embeddings / torch.sqrt(sum_mask))

        if all_layer_embeddings is not None and self.pooling_mode_mean_last_2_tokens:  # avg of last 2 layers
            input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            sum_mask = input_mask_expanded.sum(1)
            sum_mask = torch.clamp(sum_mask, min=1e-9)

            token_embeddings_last1 = all_layer_embeddings[-1]
            sum_embeddings_last1 = torch.sum(token_embeddings_last1 * input_mask_expanded, 1)
            sum_embeddings_last1 = sum_embeddings_last1 / sum_mask

            token_embeddings_last2 = all_layer_embeddings["all_layer_embeddings"][-2]
            sum_embeddings_last2 = torch.sum(token_embeddings_last2 * input_mask_expanded, 1)
            sum_embeddings_last2 = sum_embeddings_last2 / sum_mask

            output_vectors.append((sum_embeddings_last1 + sum_embeddings_last2) / 2)

        if all_layer_embeddings is not None and self.pooling_mode_mean_first_last_tokens:  # avg of the first and the last layers
            input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
            sum_mask = input_mask_expanded.sum(1)
            sum_mask = torch.clamp(sum_mask, min=1e-9)

            token_embeddings_first = all_layer_embeddings[0]
            sum_embeddings_first = torch.sum(token_embeddings_first * input_mask_expanded, 1)
            sum_embeddings_first = sum_embeddings_first / sum_mask

            token_embeddings_last = all_layer_embeddings[-1]
            sum_embeddings_last = torch.sum(token_embeddings_last * input_mask_expanded, 1)
            sum_embeddings_last = sum_embeddings_last / sum_mask

            output_vectors.append((sum_embeddings_first + sum_embeddings_last) / 2)

        output_vector = torch.cat(output_vectors, 1)
        assert output_vector.size(1)==self.pooling_output_dimension
        return output_vector

    def get_sentence_embedding_dimension(self):
        return self.pooling_output_dimension

    def get_config_dict(self):
        return {key: self.__dict__[key] for key in self.config_keys}

    def save(self, output_path):
        os.makedirs(output_path,exist_ok=True)
        with open(os.path.join(output_path, 'config.json'), 'w') as fOut:
            json.dump(self.get_config_dict(), fOut, indent=2)
        torch.save(self.state_dict(),os.path.join(output_path, 'pytorch_model.bin'))
        
    @staticmethod
    def load(input_path):
        with open(os.path.join(input_path, 'config.json')) as fIn:
            config = json.load(fIn)

        return Pooling(**config)


# -*- coding: UTF-8 -*-
################################################################################
#
#   Copyright (c) 2019  Baidu.com, Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################
"""
Setup script.
Authors: xianghuisun(2357094733@qq.com)
Date:    2021/8/20 00:00:01
"""
import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
    name="nlp-basictasks",
    version="0.1.0",
    author="xianghuisun",
    author_email="2357094733@qq.com",
    description="a simple framework that can quickly build some basic NLP tasks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/xianghuisun/nlp-basictasks",
    # packages=setuptools.find_packages(),
    #packages = ['basictasks', 
    #    'basictasks.readers', 
    #    'basictasks.modules', 
    #    'basictasks.heads',
    #    'basictasks.tasks', 
    #    'basictasks.evaluation'],
    #package_dir={'basictasks':'./basictasks',
    #             'basictasks.readers':'./basictasks/readers',
    #             'basictasks.modules':'./basictasks/modules',
    #             'basictasks.heads':'./basictasks/heads',
    #             'basictasks.tasks': './basictasks/tasks',
    #             'basictasks.evaluation': './basictasks/evaluation'},
    #package_dir={"": "nlp_basictasks"},
    #packages=setuptools.find_packages(where="nlp_basictasks"),
    packages=setuptools.find_packages(),
    platforms = "any",
    license='Apache 2.0',
    classifiers = [
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
          ],
    python_requires='>=3.0',
    install_requires = [
        'torch>=1.6.0',
        'torchvision',
        'tqdm',
        'scipy',
        'scikit-learn',
        'seqeval'
    ]
)
